-- MODULE:
--   AppState
--
-- PURPOSE:
--   Handles application state. Also contains drawing code.
--
-- AUTHOR:
--   Stephen Thompson <stephen@solarflare.org.uk>
--
-- COPYRIGHT:
--   Copyright (C) Stephen Thompson, 2011.
--
--   This file is part of Knights Map Editor.
--
--   Knights Map Editor is free software: you can redistribute it
--   and/or modify it under the terms of the GNU General Public
--   License as published by the Free Software Foundation, either
--   version 3 of the License, or (at your option) any later version.
--
--   Knights Map Editor is distributed in the hope that it will be
--   useful, but WITHOUT ANY WARRANTY; without even the implied
--   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
--   See the GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with Knights Map Editor. If not, see
--   <http://www.gnu.org/licenses/>.


module AppState 
    (
     AppStateRef
    ,EditMode(..)
    ,MainWindowWidgets(..) -- used by newAppState 

-- Initialization, load, save
    ,newAppState           -- See below
    ,setKnightsPaths       -- AppStateRef -> String -> String -> IO ()
    ,getKnightsPaths       -- AppStateRef -> IO (String, String)
    ,loadRoomsFile         -- AppStateRef -> String -> IO ()
    ,saveRoomsFile         -- AppStateRef -> IO ()
    ,saveRoomsFileAs       -- AppStateRef -> String -> IO ()
    ,isUnsaved             -- AppStateRef -> IO Bool
    ,resetAppState         -- AppStateRef -> IO ()    -- "closes" current file, clears all data.

-- Testing (running the Knights exe)
    ,testCurrentSegment    -- AppStateRef -> IO ()
    ,cleanUpTempFiles      -- AppStateRef -> Bool -> IO ()

-- State reading functions
    ,getEditMode           -- AppStateRef -> IO EditMode
    ,getMouseButton        -- AppStateRef -> IO (Maybe MouseButton)
    ,getMouseHover         -- AppStateRef -> IO (Maybe (Int,Int)) -- tile coords
    ,getLeftMouseTile      -- AppStateRef -> IO (Maybe Tile)
    ,getRightMouseTile     -- AppStateRef -> IO (Maybe Tile)
    ,isSegSelected         -- AppStateRef -> IO Bool
    ,getSwitch             -- AppStateRef -> Int -> Int -> IO (Maybe KValue)   -- returns the action only.
    ,isFileNameEmpty       -- AppStateRef -> IO Bool. True if this is a "new" file so have to use Save As not Save.

-- State writing functions (undo events NOT generated / "unsaved" state NOT changed)
    ,setMouseHover         -- AppStateRef -> Int -> Int -> IO ()
    ,setShowTileBrush      -- AppStateRef -> Bool -> IO ()
    ,setMouseButton        -- AppStateRef -> Maybe MouseButton -> IO ()
    ,setSelectedTile       -- Bool -> AppStateRef -> Int -> IO ()
    ,selectTileFromMap     -- Bool -> AppStateRef -> Int -> Int -> IO ()
    ,setCurrentSegment     -- AppStateRef -> String -> IO ()
    ,setEditMode           -- AppStateRef -> EditMode -> IO ()
    ,startMouseDrag        -- AppStateRef -> Int -> Int -> IO ()
    ,cancelMouseDrag       -- AppStateRef -> IO ()

-- State writing functions (undo events generated / "unsaved" state set)
    ,writeTileToCurrSeg    -- AppStateRef -> Tile -> Int -> Int -> IO ()
    ,setSwitch             -- AppStateRef -> Int -> Int -> KValue -> IO ()   -- takes x, y, action.
    ,rmSwitch              -- AppStateRef -> Int -> Int -> IO ()
    ,updateSegName         -- AppStateRef -> IO ()   -- Copies seg name from text-entry into the AppState. Renames if needed.
    ,updateSegCategory     -- AppStateRef -> IO ()   -- Ditto for seg category
    ,setRoomFromDrag       -- AppStateRef -> IO ()
    ,deleteRoomAt          -- AppStateRef -> Int -> Int -> IO ()
    ,addNewSegment         -- AppStateRef -> IO ()
    ,deleteCurrSegment     -- AppStateRef -> IO ()
     
-- Undo/Redo
    ,undo                  -- AppStateRef -> IO ()
    ,redo                  -- AppStateRef -> IO ()
     
-- Converting mouse positions
    ,mouseToTileCoords          -- Int -> Int -> (Int,Int)
    ,mouseToSegmentName         -- AppStateRef -> Int -> Int -> IO (Maybe String)
    ,mouseToTileNum             -- DrawingArea -> Int -> Int -> IO Int

-- Drawing
    ,drawThumbs                 -- AppStateRef -> Rectangle -> IO ()
    ,updateSegmentsScrollbar    -- AppStateRef -> IO ()
    ,drawMap                    -- AppStateRef -> Rectangle -> IO ()
    ,drawSelectedTile           -- AppStateRef -> (AppState -> DrawingArea) -> (AppState -> Tile) -> Rectangle -> IO ()
    ,drawTilePalette            -- AppStateRef -> DrawingArea -> Rectangle -> IO ()
    )

where

import DivRoundUp
import GfxLoader
import KConfig
import MapData

import Control.Exception (bracket)
import Control.Monad (when)
import Data.Char
import Data.IORef
import Data.List
import Data.Map (Map)
import qualified Data.Map as Map
import Data.Maybe
import Graphics.UI.Gtk
import Graphics.UI.Gtk.Gdk.Events
import Graphics.UI.Gtk.Gdk.GC
import System.Directory
import System.FilePath
import System.IO
import System.IO.Error
import System.Process


-- | The editor has three modes: paint (place tiles), switches (edit
-- switches and traps), rooms (place room outlines).
data EditMode = PaintMode | SwitchesMode | RoomsMode deriving Eq

-- | Holds widgets of interest from the main window
data MainWindowWidgets = MainWindowWidgets { segmentsDrawingArea :: DrawingArea
                                           , segmentsScrollbar :: VScrollbar
                                           , mapDrawingArea :: DrawingArea
                                           , leftTileDrawingArea :: DrawingArea
                                           , rightTileDrawingArea :: DrawingArea
                                           , tilePalette :: DrawingArea
                                           , segNameEntry :: Entry
                                           , segCategoryEntry :: Entry
                                           }

-- | State that gets restored during an "undo" operation.
-- (Currently segments map and current segment information, and "unsaved" flag.)
type UndoState = (Map String Segment, Maybe String, Bool)

-- | All application state is stored in the AppState record.
data AppState = AppState {
     -- gfx loader
     asGfxLoaderRef :: GfxLoaderRef

     -- widgets etc
    ,widgets :: MainWindowWidgets

    -- map/room data
    ,segments :: Map String Segment
    ,tiles :: [Tile]
    ,thumbnails :: Map String Pixbuf   -- cached thumbnails

    -- the knights_data directory (used for "include" files) 
    -- and knights exe path (will be used for testing maps)
    ,knightsDataDir :: String 
    ,knightsExePath :: String
     
    -- the currently opened file
    ,currentFileName :: String

    -- gui state 
    ,currentSegment :: Maybe String
    ,leftMouseTile :: Maybe Tile
    ,rightMouseTile :: Maybe Tile
    ,mouseHover :: Maybe (Int,Int)
    ,showTileBrush :: Bool
    ,dragStartPos :: Maybe (Int,Int)
    ,mouseButton :: Maybe MouseButton
    ,switchesListStore :: ListStore String
    ,editMode :: EditMode
     
    -- undo, and "unsaved" flag
    ,undoStack :: [UndoState]
    ,redoStack :: [UndoState]
    ,unsavedFlag :: Bool

    -- temp filenames & process handles for Knights sub-processes
    ,knightsProcesses :: [(FilePath, Maybe ProcessHandle)]
}

type AppStateRef = IORef AppState



-- Initialization

-- | Create the initial application state
newAppState :: MainWindowWidgets -> TreeView -> IO AppStateRef
newAppState wdgts switchesTreeView = do
  gfxLoaderRef <- newGfxLoader ""
  listStore <- listStoreNew []
  setupSwitchesTreeView listStore switchesTreeView
  result <- newIORef $ AppState { asGfxLoaderRef = gfxLoaderRef,
                                  widgets = wdgts,
                                  segments = Map.empty,
                                  thumbnails = Map.empty,
                                  knightsDataDir = "",
                                  knightsExePath = "",
                                  currentFileName = "",
                                  tiles = [],
                                  currentSegment = Nothing,
                                  leftMouseTile = Nothing,
                                  rightMouseTile = Nothing,
                                  mouseHover = Nothing,
                                  showTileBrush = True,
                                  dragStartPos = Nothing,
                                  mouseButton = Nothing,
                                  switchesListStore = listStore,
                                  editMode = PaintMode,
                                  undoStack = [],
                                  redoStack = [],
                                  unsavedFlag = False,
                                  knightsProcesses = []
                                }
  return result

-- | Set knights data dir and exe path. NOTE: changes to kts data dir are only applied when a new room file 
-- is opened, or resetAppState is called.
setKnightsPaths :: AppStateRef -> String -> String -> IO ()
setKnightsPaths appStateRef ktsDataDir ktsExePath = do
  gfxLoaderRef <- newGfxLoader ktsDataDir
  modifyIORef appStateRef (\as -> as { knightsDataDir = ktsDataDir, knightsExePath = ktsExePath, asGfxLoaderRef = gfxLoaderRef })
  loadDefaultTiles appStateRef

getKnightsPaths :: AppStateRef -> IO (String, String)
getKnightsPaths appStateRef = do
  appState <- readIORef appStateRef
  return (knightsDataDir appState, knightsExePath appState)

-- | Load tiles and rooms from a knights_rooms file. Any existing tiles/rooms are lost.
loadRoomsFile :: AppStateRef -> String -> IO ()
loadRoomsFile appStateRef filename = do
  oldAppState <- readIORef appStateRef
  let path = knightsDataDir oldAppState
  let ldr f = readFile (path `combine` f)
  (Right kcfg) <- loadKConfig filename ldr   -- TODO handle parse (or file not found) errors more gracefully
  let Just knights_rooms = kTableLookup "knights_rooms" kcfg
  let (namesAndSegments, tilesList) = loadSegmentsAndTiles knights_rooms
      segsMap = Map.fromList namesAndSegments
  modifyIORef appStateRef (\as -> as { segments = segsMap,
                                       thumbnails = Map.empty,
                                       tiles = tilesList,
                                       currentSegment = Nothing,
                                       leftMouseTile = Nothing,
                                       rightMouseTile = Nothing,
                                       mouseHover = Nothing,
                                       showTileBrush = True,
                                       dragStartPos = Nothing,
                                       mouseButton = Nothing,
                                       currentFileName = filename,
                                       undoStack = [],   -- Clear out undo when loading a new file.
                                       redoStack = [],
                                       unsavedFlag = False
                                     } )
  updateSegmentsScrollbar appStateRef
  appState <- readIORef appStateRef
  redrawEverything appState

-- | Try to load the default tiles from knights_rooms.txt. If fails, just return an empty list of tiles.
loadDefaultTiles :: AppStateRef -> IO ()
loadDefaultTiles appStateRef = do
  -- TODO: This could be more robust. If there is a parse error (e.g. irrefutable pattern match failure during parsing)
  -- then it just dies -- the exception isn't caught by the try block. We probably should fix the kconfig parser not to 
  -- have irrefutable pattern matches I suppose, although could also fix it by catching the exception here (using
  -- Control.Exception).
  x <- try $ do
    oldAppState <- readIORef appStateRef
    let ldr f = readFile ((knightsDataDir oldAppState) `combine` f)
    result <- loadKConfig "knights_rooms.txt" ldr
    case result of 
      Left _ -> return []
      Right kcfg -> case kTableLookup "knights_rooms" kcfg of
        Nothing -> return []
        Just knights_rooms -> do
          let (_, tilesList) = loadSegmentsAndTiles knights_rooms
          return tilesList
  case x of
    Left _ -> return ()
    Right tilesList -> modifyIORef appStateRef (\as -> as { tiles = tilesList })
        

-- | True if the file has been changed since last save (or initial load).
isUnsaved :: AppStateRef -> IO Bool
isUnsaved asr = do
  as <- readIORef asr
  return (unsavedFlag as)

-- | Save to the current filename. (Current filename must be nonempty otherwise this will crash!)
-- NOTE: This does not save if the file is not "unsaved"!
saveRoomsFile :: AppStateRef -> IO ()
saveRoomsFile asr = do
  unsaved <- isUnsaved asr
  when unsaved $ do
    as <- readIORef asr
    saveRoomsFileAs asr (currentFileName as)

-- | Save to the given filename. The current filename is updated to the given filename.
-- NOTE: This always writes, even if the file is not "unsaved".
saveRoomsFileAs :: AppStateRef -> String -> IO ()
saveRoomsFileAs asr filename = do
  updateSegProperties asr
  as <- readIORef asr
  let kcfg = saveSegmentsAndTiles (Map.toList (segments as)) (tiles as)
  withFile filename WriteMode (\h -> hPutStrLn h (prettyPrintKnightsRooms kcfg))
  writeIORef asr (as { currentFileName = filename, unsavedFlag = False }) -- Update the current filename, and clear unsaved flag.

-- | Test Current Segment in the Knights exe
testCurrentSegment :: AppStateRef -> IO ()
testCurrentSegment asr = do
  updateSegProperties asr
  as <- readIORef asr
  case currentSegment as of
    Nothing -> return ()
    Just segName -> do
      let currSeg = fromJust $ Map.lookup segName (segments as)
      let kcfg = saveSegmentsAndTiles [(segName,currSeg { segCategory="" })] (tiles as)
      
      -- Create a temporary file with the selected segment and other config stuff.
      tempDir <- getTemporaryDirectory
      bracket (openTempFile tempDir "KnightsTest.txt")   -- create the temp file.
              (\(tempFileName, h) -> do                  -- close the temp file, and store its name, afterwards.
                 try (hClose h)
                 writeIORef asr (as{ knightsProcesses = (tempFileName,Nothing) : (knightsProcesses as) } ))
              (\(tempFileName, h) -> saveTempKnightsConfig h kcfg)        -- Write data to the temp file.

      -- Read the temp file name back from the updated appstate
      as2 <- readIORef asr
      let tempFileName = (fst $ head $ knightsProcesses as2)

      -- Launch the sub process, and store the process handle back into the appstate
      let knightsCmd = (knightsExePath as) ++ " -w -a -c " ++ tempFileName
      putStrLn knightsCmd
      hFlush stdout
      (_, _, _, proc) <- createProcess (shell knightsCmd)
      writeIORef asr (as2{ knightsProcesses = (tempFileName,Just proc) : (tail $ knightsProcesses as2)})

-- | Helper function used by testCurrentSegment
saveTempKnightsConfig :: Handle -> KValue -> IO ()
saveTempKnightsConfig h kcfg = do
  hPutStrLn h $ unlines ["include \"main_config.txt\""
                        ,"MAIN_MENU=["
                        ,"DungeonStuff(\"chest\", 80, ig_pot, 3)"
                        ,"DungeonStuff(\"small_table\", 75, Random(ig_pot:2, ig_table:1), 1)"
                        ,"DungeonStuff(\"table\", 67, ig_table, 1)"
                        ,"DungeonLayout(d_tiny)"
                        ,"DungeonPremapped()"
                        ,"DungeonHome(\"random_respawn\")"
                        ,"DungeonKeys(i_lockpicks, i_key1)"
                        ,"DungeonStuff(\"barrel\", 100/(8-3), ig_trap, -1)"
                        ,"DungeonStuff(\"floor\", 100/(18-3*3), ig_big)"
                        ,"DungeonGear(i_lockpicks, [1:100])"
                        ,"{"
                        ,"id=\"quest\""
                        ,"title=\"Quest\""
                        ,"values=["
                        ,"\"Map Editor Test\""
                        ,"]"
                        ,"}"
                        ,"]"
                        ,"MAIN_MENU_SPACE=[]"
                        ,"QUEST_DESCRIPTIONS=[\"\"]"
                        ]
  hPutStrLn h $ prettyPrintKnightsRooms kcfg

-- | Delete all temporary files for processes that have exited
-- (If 'force' is True then try to delete even if the process hasn't exited yet.)
cleanUpTempFiles :: AppStateRef -> Bool -> IO ()
cleanUpTempFiles asr force = do
  as <- readIORef asr
  let procs = knightsProcesses as
  newProcs' <- mapM (cleanTempFile force) procs
  let newProcs = filter (isJust . snd) newProcs'
  writeIORef asr (as { knightsProcesses = newProcs })

cleanTempFile :: Bool -> (FilePath, Maybe ProcessHandle) -> IO (FilePath, Maybe ProcessHandle)
cleanTempFile True  (path, _) = try (removeFile path) >> return (path, Nothing)
cleanTempFile False (path, Nothing) = cleanTempFile True (path, Nothing)
cleanTempFile False (path, Just proc) = do
  x <- getProcessExitCode proc
  case x of 
    Nothing -> return (path, Just proc)  -- Still running. Don't try to delete.
    Just _ -> cleanTempFile True (path, Nothing)


-- | Clears all rooms, tiles etc from the current state.
resetAppState :: AppStateRef -> IO ()
resetAppState appStateRef = do
  modifyIORef appStateRef (\as -> as { segments = Map.empty
                                     , thumbnails = Map.empty
                                     , tiles = []
                                     , currentSegment = Nothing
                                     , leftMouseTile = Nothing
                                     , rightMouseTile = Nothing
                                     , mouseHover = Nothing
                                     , showTileBrush = True
                                     , dragStartPos = Nothing
                                     , mouseButton = Nothing
                                     , undoStack = []  
                                     , redoStack = []
                                     , unsavedFlag = False } )
  updateSegmentsScrollbar appStateRef
  loadDefaultTiles appStateRef
  appState <- readIORef appStateRef
  redrawEverything appState

-- | Set up the tree view for the switches (set list store, cell renderers etc)
-- Called during initialization
setupSwitchesTreeView :: ListStore String -> TreeView -> IO ()
setupSwitchesTreeView model treeView = do
  treeViewSetModel treeView model

  renderer <- cellRendererTextNew
  col <- treeViewColumnNew
  treeViewColumnPackStart col renderer True
  cellLayoutSetAttributes col renderer model $ \row -> [ cellText := row ]
  treeViewAppendColumn treeView col

  return ()
  



-- Functions to make stuff get redrawn after various updates

-- | Redraws everything that might have changed e.g. after loading a new rooms file.
redrawEverything :: AppState -> IO ()
redrawEverything appState = do
  redrawAllTiles appState
  redrawAllSegs appState
  redrawSwitchesList appState
  redrawTilePalette appState
  rereadSegProperties appState

-- | Schedule a redraw of the segments (thumbnails) area
-- TODO: would be better to redraw only what we need to...
redrawAllSegs :: AppState -> IO ()
redrawAllSegs appState = do
  let segmentsArea = segmentsDrawingArea (widgets appState)
  widgetQueueDraw segmentsArea

-- | Redraw one particular tile on the main map area
redrawTile :: AppState -> Int -> Int -> IO ()
redrawTile appState x y = do
  let (dx,dy,dw,dh) = tileToMouseRect x y
  widgetQueueDrawArea (mapDrawingArea $ widgets appState) (dx-1) (dy-1) (dw+2) (dh+2)

-- | Redraw the entire main map area
redrawAllTiles :: AppState -> IO ()
redrawAllTiles appState = widgetQueueDraw (mapDrawingArea $ widgets appState)

-- | Recompute the switches list model and redraw the switch listview
redrawSwitchesList :: AppState -> IO ()
redrawSwitchesList as = do
  let listStore = switchesListStore as
  listStoreClear listStore
  case getSegWithName as of 
    Nothing -> return ()
    Just (seg, _) -> 
        let switchesList = Map.toList $ segSwitches seg
            strings = map (\((x,y),a) -> "(" ++ (show x) ++ "," ++ (show y) ++ ") " ++ (show a)) switchesList
        in mapM_ (listStoreAppend listStore) strings

-- | Gets current drag rectangle in form (x, y, width, height) (in tile coords)
getDragRectTileCoords :: AppState -> Maybe (Int,Int,Int,Int)
getDragRectTileCoords as =
    case (dragStartPos as, mouseHover as) of
      (Just (x1,y1), Just (x2,y2)) ->
          let xlow = min x1 x2
              xhigh = max x1 x2
              ylow = min y1 y2
              yhigh = max y1 y2
          in Just (xlow, ylow, xhigh - xlow + 1, yhigh - ylow + 1)
      _ -> Nothing

-- | Returns (x,y,width,height) (in pixels) of current drag rect, if any.
getDragRectPixelCoords :: AppState -> Maybe (Int,Int,Int,Int)
getDragRectPixelCoords as = do -- Maybe monad
  (tx,ty,tw,th) <- getDragRectTileCoords as
  let (px1,py1,_,_) = tileToMouseRect tx ty
      (px2,py2,pw2,ph2) = tileToMouseRect (tx+tw-1) (ty+th-1)
  return (px1, py1, px2-px1+pw2, py2-py1+ph2)


-- | Schedules a redraw of the current drag rectangle
redrawDragRectangle :: AppStateRef -> IO ()
redrawDragRectangle appStateRef = do
  as <- readIORef appStateRef
  case (getDragRectPixelCoords as) of
    Nothing -> return ()
    Just (dx,dy,dw,dh) -> widgetQueueDrawArea (mapDrawingArea $ widgets as) (dx-1) (dy-1) (dw+2) (dh+2)

-- | Schedules a redraw of the tile palette
redrawTilePalette :: AppState -> IO ()
redrawTilePalette appState = widgetQueueDraw (tilePalette $ widgets appState)


-- Various helper functions

-- | Get currently selected segment and its name
getSegWithName :: AppState -> Maybe (Segment, String)
getSegWithName as = 
  case currentSegment as of 
    Nothing -> Nothing
    Just segName -> Just (fromJust $ Map.lookup segName (segments as), segName)

-- | Updates (or adds) a segment in the appstate. Also flushes thumbnail from the cache.
-- Returns a new appstate. Caller should take care of any redrawing needed.
updateSegment :: String -> Segment -> AppState -> AppState
updateSegment name seg as = as { segments = Map.insert name seg (segments as),
                                 thumbnails = Map.delete name (thumbnails as) }


-- Undo handling

currUndoState :: AppState -> UndoState
currUndoState as = (segments as, currentSegment as, unsavedFlag as)

-- | Push the current state onto the top of the "undo stack"
-- Also sets the "unsaved" flag (in the new appstate).
addUndo :: AppState -> AppState
addUndo as = as { undoStack = (currUndoState as) : (undoStack as), redoStack = [], unsavedFlag = True }

-- | Do an "undo"
undo :: AppStateRef -> IO ()
undo appStateRef = do
  updateSegProperties appStateRef
  oldAppState <- readIORef appStateRef
  case (undoStack oldAppState) of
    [] -> return ()
    (x : xs) -> doUndoRedo appStateRef x xs ((currUndoState oldAppState) : (redoStack oldAppState))
    
-- | Do a "redo"
redo :: AppStateRef -> IO ()
redo appStateRef = do
  updateSegProperties appStateRef
  oldAppState <- readIORef appStateRef
  case (redoStack oldAppState) of
    [] -> return ()
    (x : xs) -> doUndoRedo appStateRef x ((currUndoState oldAppState) : (undoStack oldAppState)) xs

doUndoRedo :: AppStateRef -> UndoState -> [UndoState] -> [UndoState] -> IO ()
doUndoRedo appStateRef (newSegs, newCurrSeg, newUnsavedFlag) newUndoStack newRedoStack = do
  modifyIORef appStateRef (\as -> as { segments = newSegs
                                     , currentSegment = newCurrSeg
                                     , thumbnails = Map.empty  -- probably overkill
                                     , mouseHover = Nothing
                                     , showTileBrush = False
                                     , dragStartPos = Nothing
                                     , mouseButton = Nothing
                                     , undoStack = newUndoStack
                                     , redoStack = newRedoStack
                                     , unsavedFlag = newUnsavedFlag
                                     })
  newAppState <- readIORef appStateRef
  redrawEverything newAppState  -- Overkill, but it will do for now.



-- State reading functions

getEditMode a = readIORef a >>= (return . editMode)
getMouseButton a = readIORef a >>= (return . mouseButton)
getMouseHover a = readIORef a >>= (return . mouseHover)
getLeftMouseTile a = readIORef a >>= (return . leftMouseTile)
getRightMouseTile a = readIORef a >>= (return . rightMouseTile)

-- | Is a segment currently selected?
isSegSelected :: AppStateRef -> IO Bool
isSegSelected asr = do
  as <- readIORef asr
  case getSegWithName as of
    Nothing -> return False
    _ -> return True

-- | Get the switch at the given position, or Nothing if no switch is defined here.
getSwitch :: AppStateRef -> Int -> Int -> IO (Maybe KValue)
getSwitch asr x y = do  
  as <- readIORef asr
  return $ do   -- Maybe monad
    (seg, _) <- getSegWithName as
    Map.lookup (x,y) (segSwitches seg)

-- | True if this is a "new" file so have to use Save As not Save
isFileNameEmpty :: AppStateRef -> IO Bool
isFileNameEmpty asr = do
  as <- readIORef asr
  return (null (currentFileName as))
  

-- State writing functions
-- These will queue redraws as necessary

-- | Updates a tile in the current segment.
-- Adds an undo step.
-- (Does nothing if the coords are out of bounds.)
writeTileToCurrSeg :: AppStateRef -> Tile -> Int -> Int -> IO ()
writeTileToCurrSeg appStateRef tile x y = do
  appState <- readIORef appStateRef
  case getSegWithName appState of 
    Nothing -> return ()
    Just (seg, segName) -> do
      if (x < 0 || x >= (segWidth seg) || y < 0 || y >= (segHeight seg)
         || (tileNumber (segGetTile seg (x,y))) == (tileNumber tile)) then return ()
       else do
         let baseState = addUndo appState
         let newSeg = segSetTile seg (x,y) tile
             newAppState = updateSegment segName newSeg baseState
         writeIORef appStateRef newAppState
         redrawTile newAppState x y
         redrawAllSegs newAppState

-- | Update the mouse hover position
setMouseHover :: AppStateRef -> Maybe (Int,Int) -> IO ()
setMouseHover appStateRef newHoverPos = do
  -- Get the old AppState
  oldAppState <- readIORef appStateRef
  -- Update the AppState
  let newAppState = oldAppState { mouseHover = newHoverPos }
  writeIORef appStateRef newAppState
  -- Schedule a redraw of both the old and the new positions
  case (mouseHover oldAppState) of
    Nothing -> return ()
    Just (oldTX, oldTY) -> redrawTile newAppState oldTX oldTY
  case newHoverPos of
    Nothing -> return ()
    Just (newTX, newTY) -> redrawTile newAppState newTX newTY
  -- Do a full redraw - temporary code
  -- TODO: Should only redraw what needs to be i.e. the tiles around the edge of drag rect
  -- and only when dragging, in rooms mode
  redrawAllTiles newAppState
  -- This makes sure the next motion event reaches us:
  win <- widgetGetDrawWindow (mapDrawingArea $ widgets newAppState)
  drawWindowGetPointer win
  return ()
  
-- | Set whether "tile brush" (ie image of selected tile under mouse ptr) is drawn
setShowTileBrush :: AppStateRef -> Bool -> IO ()
setShowTileBrush appStateRef setting = do
  oldAppState <- readIORef appStateRef
  let newAppState = oldAppState { showTileBrush = setting }
  writeIORef appStateRef newAppState
  case (mouseHover newAppState) of
    Nothing -> return ()
    Just (tx,ty) -> redrawTile newAppState tx ty

-- | Update the mouse button state
setMouseButton :: AppStateRef -> Maybe MouseButton -> IO ()
setMouseButton appStateRef newSetting = modifyIORef appStateRef (\as -> as { mouseButton = newSetting })

-- | Remove the switch at the given position (if there is one). Adds an undo step.
rmSwitch :: AppStateRef -> Int -> Int -> IO ()
rmSwitch asr x y = do
  as <- readIORef asr
  case (getSegWithName as) of
    Nothing -> return ()
    Just (seg, name) -> do
      let baseState = addUndo as
      let oldSwitches = segSwitches seg
          newSwitches = Map.delete (x,y) oldSwitches
          newSeg = seg { segSwitches = newSwitches }
          newAppState = updateSegment name newSeg baseState
      writeIORef asr newAppState
      redrawTile newAppState x y
      redrawSwitchesList newAppState

-- | Add the given switch, replacing any existing switch at the same position if necessary.
-- Adds an undo step.
setSwitch :: AppStateRef -> Int -> Int -> KValue -> IO ()
setSwitch asr sx sy sa = do
  as <- readIORef asr
  case (getSegWithName as) of
    Nothing -> return ()
    Just (seg, name) -> do
      let baseState = addUndo as
      -- update the appstate itself.
      let oldSwitches = segSwitches seg
          newSwitches = Map.insert (sx,sy) sa oldSwitches
          newSeg = seg { segSwitches = newSwitches }
          newAppState = updateSegment name newSeg baseState
      writeIORef asr newAppState
      -- redraw the affected tile
      redrawTile newAppState sx sy
      -- redraw the switches listview.
      redrawSwitchesList newAppState

-- | Change the currently selected segment
setCurrentSegment :: AppStateRef -> String -> IO ()
setCurrentSegment appStateRef segName = do
  updateSegProperties appStateRef
  asOld <- readIORef appStateRef
  when (isJust (Map.lookup segName (segments asOld)))  -- check that updateSegProperties didn't zap that name
    (doSetCurrentSegment appStateRef (Just segName))
    
doSetCurrentSegment :: AppStateRef -> Maybe String -> IO ()
doSetCurrentSegment asr seg = do
  modifyIORef asr (\as -> as { currentSegment = seg })
  as <- readIORef asr
  redrawAllTiles as
  redrawSwitchesList as
  rereadSegProperties as
  
  
    
-- | Reread name, category of the current segment and write these to the text entry widgets
rereadSegProperties :: AppState -> IO ()
rereadSegProperties as = do
  case (currentSegment as) of
    Nothing -> do
      entrySetText (segNameEntry $ widgets as) ""
      entrySetText (segCategoryEntry $ widgets as) ""
    Just name -> do
      let seg = fromJust $ Map.lookup name (segments as)
          category = segCategory seg
      entrySetText (segNameEntry $ widgets as) name
      entrySetText (segCategoryEntry $ widgets as) category

-- | Rename a proposed seg name if it clashes with an existing name
renameSegment :: AppState -> String -> String
renameSegment as requestedName = 
  case (Map.lookup requestedName (segments as)) of 
    Nothing -> requestedName
    Just _ -> renameSegment as (makeNewName requestedName)
    
makeNewName :: String -> String
makeNewName foo =
  case (findIndex (=='_') (reverse foo)) of
    Nothing -> foo ++ "_1"
    Just i -> let pos = (length foo) - i   -- 1 to (length foo)
                  (prefix,rest) = splitAt pos foo
                  restAsInt :: Int
                  restAsInt = read rest
               in if (not (null rest) && (and (map isDigit rest))) then 
                    prefix ++ show (restAsInt + 1)
                   else foo ++ "_1"
                 
-- | Force all segment properties (name, category) to be updated
updateSegProperties :: AppStateRef -> IO ()
updateSegProperties = updateSegName >> updateSegCategory

-- | Rename the currently selected segment (if necessary). Adds an undo step if the name was changed.
updateSegName :: AppStateRef -> IO ()
updateSegName appStateRef = do
  as <- readIORef appStateRef
  case currentSegment as of 
    Nothing -> return ()
    Just currName -> do
      requestedName <- entryGetText (segNameEntry $ widgets as)
      let newName = renameSegment as requestedName
      when (requestedName /= currName) $ do
        let baseState = addUndo as
        let currSeg = fromJust $ Map.lookup currName (segments baseState)
        let newSegments = Map.insert newName currSeg (Map.delete currName (segments baseState))
        writeIORef appStateRef (baseState { segments = newSegments, 
                                            currentSegment = Just newName, 
                                            thumbnails = Map.delete currName (Map.delete newName (thumbnails as))
                                   })
        entrySetText (segNameEntry $ widgets baseState) newName
        redrawAllSegs baseState

-- | Update the seg category from the text field (if it is different from the current category).
-- Adds an undo step.
updateSegCategory :: AppStateRef -> IO ()
updateSegCategory appStateRef = do
  as <- readIORef appStateRef
  case currentSegment as of 
    Nothing -> return ()
    Just currName -> do
      requestedCategory <- entryGetText (segCategoryEntry $ widgets as)
      let currSeg = fromJust $ Map.lookup currName (segments as)
      when (requestedCategory /= (segCategory currSeg)) $ do
         let baseState = addUndo as
         let newSeg = currSeg { segCategory = requestedCategory }
         writeIORef appStateRef (updateSegment currName newSeg as)

-- | Change the current edit mode
setEditMode :: AppStateRef -> EditMode -> IO ()
setEditMode appStateRef newMode = do
  modifyIORef appStateRef (\as -> as { editMode = newMode })
  as <- readIORef appStateRef
  widgetQueueDraw (mapDrawingArea $ widgets as)

-- | Change the currently selected tile (for either left or right mouse)
setSelectedTile :: Bool -> AppStateRef -> Int -> IO ()
setSelectedTile isLeft appStateRef tileNum = do
  appState <- readIORef appStateRef
  let nTiles = length $ tiles appState
  if (tileNum < 0 || tileNum >= nTiles)
    then return ()
    else do
      let newTile = (tiles appState)!!tileNum
      doSetSelectedTile isLeft appStateRef newTile

doSetSelectedTile :: Bool -> AppStateRef -> Tile -> IO ()
doSetSelectedTile isLeft appStateRef tile = do
  appState <- readIORef appStateRef
  if isLeft then do
      widgetQueueDraw (leftTileDrawingArea $ widgets appState)
      writeIORef appStateRef $ appState { leftMouseTile = Just tile }
    else do
      widgetQueueDraw (rightTileDrawingArea $ widgets appState)
      writeIORef appStateRef $ appState { rightMouseTile = Just tile }

-- | Set selected tile to the tile at the given map pos
selectTileFromMap :: Bool -> AppStateRef -> Int -> Int -> IO ()
selectTileFromMap isLeft appStateRef tx ty = do
  appState <- readIORef appStateRef
  case currentSegment appState of
    Nothing -> return ()
    Just segName -> do
      let seg = fromJust $ Map.lookup segName (segments appState)
      let tile = segGetTile seg (tx,ty)
      doSetSelectedTile isLeft appStateRef tile


-- | Start a mouse drag (used for Rooms mode). The start position of the drag is assumed to
-- be the current "mouse hover" position.
startMouseDrag :: AppStateRef -> Int -> Int -> IO ()
startMouseDrag appStateRef tx ty = do
  modifyIORef appStateRef (\as -> as { dragStartPos = mouseHover as })
  redrawDragRectangle appStateRef

-- | Cancel any mouse drag operation in progress
cancelMouseDrag :: AppStateRef -> IO ()
cancelMouseDrag appStateRef = do
  modifyIORef appStateRef (\as -> as { dragStartPos = Nothing })
  redrawDragRectangle appStateRef

-- | Determine whether the two rooms "overlap"
-- Rooms are specified as (x,y,width,height)
roomOverlap :: (Int,Int,Int,Int) -> (Int,Int,Int,Int) -> Bool
roomOverlap (x1,y1,w1,h1) (x2,y2,w2,h2) =
    let missX = x1+w1-1 < x2 || x2+w2-1 < x1
        missY = y1+h1-1 < y2 || y2+h2-1 < y1
    in not (missX || missY)

-- | Determine whether a given point is in the "interior" of 
-- a room (i.e. not on one of the border squares)
roomInside :: (Int,Int) -> (Int,Int,Int,Int) -> Bool
roomInside (x,y) (rx,ry,rw,rh) =
    x >= rx+1 && x <= rx+rw-2 && y >= ry+1 && y <= ry+rh-2

-- | Add a room occupying the current drag rectangle, plus a one-square border around it.
-- Any existing rooms overlapping the rectangle (or a one-square border around it) 
-- will be deleted.
-- Adds an undo step.
setRoomFromDrag :: AppStateRef -> IO ()
setRoomFromDrag appStateRef = do
  oldAppState <- readIORef appStateRef
  case (getDragRectTileCoords oldAppState) of 
    Just (x,y,w,h) | w>0 && h>0 -> do
      case getSegWithName oldAppState of 
        Nothing -> return ()
        Just (seg, name) -> do
          let baseState = addUndo oldAppState
          let oldRooms = segRooms seg
              newRoom = (x+1-rmMargin,y+1-rmMargin,w+2*rmMargin,h+2*rmMargin)
              newRooms =  newRoom : filter (not . (roomOverlap (x+1,y+1,w,h))) oldRooms
              newAppState = updateSegment name (seg {segRooms = newRooms}) baseState
          writeIORef appStateRef newAppState
          redrawAllTiles newAppState
    _ -> return ()

-- | Delete any room at the given tile position. Adds an undo step.
deleteRoomAt :: AppStateRef -> Int -> Int -> IO ()
deleteRoomAt appStateRef x y = do
  oldAppState <- readIORef appStateRef
  case getSegWithName oldAppState of 
    Nothing -> return ()
    Just (seg, name) -> do
      let oldRooms = segRooms seg
          newRooms = filter (not . (roomInside (x+1,y+1))) oldRooms
      when (newRooms /= (segRooms seg)) $ do 
          let newAppState = updateSegment name (seg {segRooms = newRooms}) (addUndo oldAppState)
          writeIORef appStateRef newAppState
          redrawAllTiles newAppState

-- | Finds the tile with label "-" which is a special value meaning "this is the default floor tile".
getFloorTile :: [Tile] -> Tile
getFloorTile tiles =
  case find (\t -> tileLabel t == "-") tiles of
    Just t -> t
    Nothing -> head tiles
  

-- | Add a new segment, and select it. Size is hard coded currently. 
-- Adds undo step.
addNewSegment :: AppStateRef -> IO ()
addNewSegment asr = do
  oldAppState' <- readIORef asr
  let oldAppState = addUndo oldAppState'
  let oldSegments = segments oldAppState
      newSegName = renameSegment oldAppState "New"
      newSegment = blankSegment (getFloorTile $ tiles oldAppState)
  let newAppState = updateSegment newSegName newSegment oldAppState
  writeIORef asr newAppState
  redrawAllSegs newAppState
  setCurrentSegment asr newSegName
  scrollToCurrSeg asr
  updateSegmentsScrollbar asr
  
  
-- | Delete the currently selected segment. Adds undo step.
deleteCurrSegment :: AppStateRef -> IO ()
deleteCurrSegment asr = do
  oldAppState' <- readIORef asr
  case (currentSegment oldAppState') of
    Nothing -> return ()
    Just segName -> do
      let oldAppState = addUndo oldAppState'
      let newAppState = oldAppState { segments = Map.delete segName (segments oldAppState),
                                      thumbnails = Map.delete segName (thumbnails oldAppState) }
      writeIORef asr newAppState
      redrawAllSegs newAppState
      doSetCurrentSegment asr Nothing
      updateSegmentsScrollbar asr


--
-- DRAWING ROUTINES
--


-- Constants

-- | Width of a "small" tile (used within the segment thumbnails).
smallW = 16
-- | Height of a "small" tile.
smallH = 16

-- | Width of a "large" tile (used within the map views).
largeW = 32  -- TODO this needs to be user settable perhaps. 32x32 is more appropriate for smaller displays.
-- | Height of a "large" tile.
largeH = 32

-- | Controls drawing of room outlines
rmMargin = 1

-- | Interpolation to use
interpType = InterpNearest

-- | Left border for each thumbnail
borderLeft = 10
-- | Right border for each thumbnail
borderRight = 10
-- | Bottom border for each thumbnail
borderBottom = 20
-- | Top border for each thumbnail
borderTop = 10

-- | Number of columns of segment thumbnails
numSegCols = 1


-- Segment Thumbnails

-- | Draw the thumbnails for all segments
drawThumbs :: AppStateRef 
           -> Rectangle   -- ^ Only thumbnails intersecting this rectangle will be drawn
           -> IO ()
drawThumbs appStateRef (Rectangle dx dy dw dh) = do
  state <- readIORef appStateRef
  let drawingArea = segmentsDrawingArea (widgets state)
  if Map.null (segments state)
    then return ()
    else do
      let items = Map.assocs (segments state)     -- [(String,Segment)]
      -- Compute size of each pixbuf (incl. border)
      -- Note: we are assuming that all the pixbufs have the same size.
      let firstSeg = snd (head items)
          (iconWidth, iconHeight) = calcIconSize firstSeg
      let rows = (length items) `divRoundUp` numSegCols

      win <- widgetGetDrawWindow drawingArea

      -- Figure out current scrollbar position.
      adj <- rangeGetAdjustment (segmentsScrollbar $ widgets state)
      displacementDbl <- adjustmentGetValue adj
      let displacement = round displacementDbl

      -- Convert "items" and remove entries that fall outside the rectangle
      let rawPositions = [(x*iconWidth, y*iconHeight) | y <- [0..], x <- [0..(numSegCols-1)]]
          positions = map (\(x,y)->(x,y-displacement)) rawPositions
      let newItems = zip positions items
      let filterFunc ((x,y), (name,seg)) = not (x > (dx+dw) || (x+iconWidth) < dx || y > (dy+dh) || (y+iconHeight) < dy)
          filtered = filter filterFunc newItems

      -- Draw the filtered items
      gc <- gcNew win
      pangoContext <- widgetGetPangoContext drawingArea
      mapM_ (drawThumb appStateRef win gc pangoContext) filtered


-- | Calculate the size of one of the segment thumbnails (including its borders).
calcIconSize :: Segment -> (Int,Int)
calcIconSize seg = 
    let pbw = (segWidth seg) * smallW
        pbh = (segHeight seg) * smallH
    in (pbw + borderLeft + borderRight, 
        pbh + borderBottom + borderTop)


-- | Draw a single thumbnail
drawThumb :: AppStateRef
          -> DrawWindow    -- ^ DrawWindow to draw into
          -> GC            -- ^ Graphics context to draw into
          -> PangoContext  -- ^ Pango context for the drawing area
          -> ((Int,Int),(String,Segment))  -- ^ (x,y) to draw at, segment name, and segment contents.
          -> IO ()
drawThumb appStateRef win gc pangoContext ((x,y),(title,seg)) = do
  -- Draw the thumbnail itself
  pixbuf <- lookupThumbnail appStateRef title seg
  drawPixbuf win gc pixbuf 0 0 (x+borderLeft) (y+borderTop) (-1) (-1) RgbDitherNormal 0 0
  -- Draw the name underneath
  pangoLayout <- layoutText pangoContext title
  w <- pixbufGetWidth pixbuf
  h <- pixbufGetHeight pixbuf
  (inkSize, Rectangle _ _ logicalWidth logicalHeight) <- layoutGetPixelExtents pangoLayout
  drawLayout win gc (x+borderLeft+((w-logicalWidth)`div`2)) (y+borderTop+h+2) pangoLayout


-- | Look up a segment's thumbnail image in the AppState. If it doesn't exist, create it.
lookupThumbnail :: AppStateRef   -- ^ The application state
                -> String        -- ^ Segment name
                -> Segment       -- ^ The segment
                -> IO Pixbuf     -- ^ Returns the thumbnail image for the segment.
lookupThumbnail appStateRef title seg = do
  state <- readIORef appStateRef
  case (Map.lookup title (thumbnails state)) of
    (Just p) -> return p
    Nothing -> do
      p <- createThumbnail (asGfxLoaderRef state) seg
      let state' = state { thumbnails = Map.insert title p (thumbnails state) }
      writeIORef appStateRef state'
      return p


-- | Create the thumbnail image for a segment.
createThumbnail :: GfxLoaderRef -> Segment -> IO Pixbuf
createThumbnail gfxLoaderRef seg = do
  let segW = segWidth seg
      segH = segHeight seg

  -- 8 bpp; include alpha channel to prevent gtk warnings
  result <- pixbufNew ColorspaceRgb True 8 (smallW * segW) (smallH * segH)

  let coords = [(x,y) | y <- [0..segW-1], x <- [0..segH-1]]
  mapM_ (renderSmallTile gfxLoaderRef result seg) coords

  return result


-- | Looks up a tile at given (x,y) position within the given segment,
-- and renders a small copy of it into the pixbuf (at appropriate position).
-- Used for drawing the segment thumbnails.
renderSmallTile :: GfxLoaderRef -> Pixbuf -> Segment -> (Int,Int) -> IO ()
renderSmallTile gfxLoaderRef output seg (x,y) = do
  let tile = segGetTile seg (x,y)
  gfx <- loadTileGraphic gfxLoaderRef tile
  gfxWidth <- pixbufGetWidth gfx
  gfxHeight <- pixbufGetWidth gfx
  let dstX = smallW * x
      dstY = smallH * y
      scaleX = ((fromIntegral smallW)::Double) / ((fromIntegral (gfxWidth))::Double)
      scaleY = ((fromIntegral smallH)::Double) / ((fromIntegral (gfxHeight))::Double)
  pixbufScale gfx output dstX dstY smallW smallH (fromIntegral dstX) (fromIntegral dstY) scaleX scaleY interpType


-- | Set up the segments scrollbar
updateSegmentsScrollbar :: AppStateRef -> IO ()
updateSegmentsScrollbar appStateRef = do
  appState <- readIORef appStateRef
  let nsegs = Map.size (segments appState)
  if nsegs == 0 then return ()
   else do
    let scrollbar = segmentsScrollbar (widgets appState)
    let (width, height) = calcIconSize (snd $ head $ Map.toList $ segments appState)
    let totalHeight = height * nsegs
    adj <- rangeGetAdjustment scrollbar
    let pageHeight = 600 -- TODO set this properly
    adjustmentSetLower adj 0
    adjustmentSetPageIncrement adj (pageHeight*0.9)
    adjustmentSetPageSize adj pageHeight
    adjustmentSetStepIncrement adj ((fromIntegral height)*0.2)
    adjustmentSetUpper adj (fromIntegral totalHeight)
    rangeSetAdjustment scrollbar adj


-- | Scroll such that the current segment is visible.
scrollToCurrSeg :: AppStateRef -> IO ()
scrollToCurrSeg appStateRef = do
  appState <- readIORef appStateRef
  case currentSegment appState of 
    Nothing -> return ()
    Just segName -> do
      let segsAsList = Map.toList $ segments appState
          allSegNames = map fst segsAsList
      case findIndex (==segName) allSegNames of 
        Nothing -> return ()
        Just i -> do
          let (width, height) = calcIconSize (snd (head segsAsList))
              preferredPos = fromIntegral (height * i)
          let scrollbar = segmentsScrollbar (widgets appState)
          -- TODO could be cleverer e.g. don't move the scrollbar if the segment is already visible.
          adj <- rangeGetAdjustment scrollbar
          adjustmentSetValue adj preferredPos
          rangeSetAdjustment scrollbar adj
          redrawAllSegs appState


-- Coordinate conversion

-- | Convert mouse coords within the map drawing area to tile coords
mouseToTileCoords :: Int -> Int -> (Int,Int)
mouseToTileCoords x y = (x `div` largeW, y `div` largeH)

-- | Return rectangle (x,y,width,height) occupied by the given tile
tileToMouseRect :: Int -> Int -> (Int,Int,Int,Int)
tileToMouseRect x y = (x*largeW, y*largeH, largeW, largeH)

-- | Convert mouse coords within the segments drawing area to a segment name (or Nothing)
mouseToSegmentName :: AppStateRef -> Int -> Int -> IO (Maybe String)
mouseToSegmentName asr x y = do
  appState <- readIORef asr
  let segs = Map.toList (segments appState)
  if (null segs) 
   then return Nothing
   else do
     -- get current scrollbar position
     adj <- rangeGetAdjustment (segmentsScrollbar $ widgets appState)
     displacementDbl <- adjustmentGetValue adj
     let displacement = round displacementDbl

     -- find out which seg the coords are within
     let (iconWidth, iconHeight) = calcIconSize (snd $ head segs)
         (xx,xm) = x `divMod` iconWidth
         (yy,ym) = (y+displacement) `divMod` iconHeight
     if (xm >= borderLeft && xm <= (iconWidth-borderRight)
         && ym >= borderTop && ym <= (iconHeight-borderBottom)
         && xx >= 0 && xx < numSegCols && yy >= 0) then do
           let n = (xx + yy*numSegCols)
           if (n < (length segs)) then return $ Just $ fst (segs !! n)
                                  else return Nothing
         else return Nothing

-- | Convert mouse coords within the tile palette drawing area to a tile num
-- NOTE: result may not be in range
mouseToTileNum :: DrawingArea -> Int -> Int -> IO Int
mouseToTileNum tileArea x y = do
  (widgetWidth, _) <- widgetGetSize tileArea
  let xTiles = widgetWidth `div` largeW
  let xx = x `div` largeW
      yy = y `div` largeH
  return $ xx + yy*xTiles



-- Main Map Area

-- | Draw the map area
drawMap :: AppStateRef -> Rectangle -> IO ()
drawMap appStateRef (Rectangle dx dy dw dh) = do
  appState <- readIORef appStateRef
  let gfxLoaderRef = asGfxLoaderRef appState
      mapArea = mapDrawingArea $ widgets appState
  case (getSegWithName appState) of
    Nothing -> return ()
    Just (seg, _) -> do
      let segW = segWidth seg
          segH = segHeight seg
      let xMin = max 0 $ min (segW-1) $ dx `div` largeW
          xMax = max 0 $ min (segW-1) $ (dx+dw) `divRoundUp` largeW
          yMin = max 0 $ min (segH-1) $ dy `div` largeH
          yMax = max 0 $ min (segH-1) $ (dy+dh) `divRoundUp` largeH
      win <- widgetGetDrawWindow mapArea
      gc <- gcNew win
      pangoContext <- widgetGetPangoContext mapArea
      mapM_ (doTile gfxLoaderRef win gc pangoContext appState) 
              [(x, y, segGetTile seg (x,y)) | y <- [yMin..yMax], x <- [xMin..xMax]]
      let drag = getDragRectTileCoords appState
      when (editMode appState == RoomsMode) $ drawRooms win gc drag seg
                                   

-- | Helper function called by drawMap
doTile :: GfxLoaderRef -> DrawWindow -> GC -> PangoContext -> AppState -> (Int,Int,Tile) -> IO ()
doTile gfxLoaderRef win gc pangoContext appState (x,y,tile) = do
  let theTile = 
        case (mouseHover appState, mouseButton appState) of
          -- If mouse is hovering over a tile, and button is released, 
          -- then draw the "hovering" tile
          -- Otherwise, draw the tile from the map ('tile')
          (Just (xx,yy), Nothing) | x == xx && y == yy && (editMode appState == PaintMode) 
                                    && (showTileBrush appState) -> 
                                      case (leftMouseTile appState) of
                                        Just t -> t
                                        Nothing -> tile
          _ -> tile
  drawTile win gc pangoContext gfxLoaderRef theTile (largeW*x) (largeH*y)
  when (editMode appState == SwitchesMode) (drawSwitchBox win gc appState x y)


-- | Draw a single tile
drawTile :: DrawWindow -> GC -> PangoContext -> GfxLoaderRef -> Tile -> Int -> Int -> IO ()
drawTile win gc pangoContext gfxLoaderRef tile x y = do
  gfx <- loadTileGraphic gfxLoaderRef tile
  gfx' <- pixbufScaleSimple gfx largeW largeH interpType
  drawPixbuf win gc gfx' 0 0 x y (-1) (-1) RgbDitherNormal 0 0
  let lab = tileLabel tile
  if (lab /= "" && lab /= "-") then 
      do
        pangoLayout <- layoutText pangoContext lab
        drawLayoutWithColors win gc x y pangoLayout (Just $ Color 65535 65535 65535) (Just $ Color 0 0 0)
   else return ()


-- | Get the "color" of a switch at the given coords (within the current segment)
readSwitchColFromMap :: AppState -> Int -> Int -> Maybe Color
readSwitchColFromMap appState x y =
    case getSegWithName appState of 
      Nothing -> Nothing
      Just (seg, _) -> 
          case Map.lookup (x,y) (segSwitches seg) of
            Nothing -> Nothing
            Just _ -> Just $ Color 65535 0 0


-- | Draw the coloured box around a "switch" tile
drawSwitchBox :: DrawWindow -> GC -> AppState -> Int -> Int -> IO ()
drawSwitchBox win gc appState x y = do
  let colour = 
          case (mouseHover appState, mouseButton appState) of
            (Just (xx,yy), Nothing) | x==xx && y==yy 
                                      && (editMode appState == SwitchesMode) -> Just $ Color 65535 65535 65535
            _ -> readSwitchColFromMap appState x y
  let xx = largeW*x
      yy = largeH*y  
  case colour of
    Just c -> do
      gv <- gcGetValues gc
      gcSetValues gc $ gv { foreground = c, lineWidth = 3 }
      drawRectangle win gc False xx yy (largeW-1) (largeH-1)
    _ -> return ()


-- | Another helper function called by drawMap -- draws room outlines.
drawRooms :: DrawWindow -> GC -> Maybe (Int,Int,Int,Int) -> Segment -> IO ()
drawRooms win gc drag seg = do
  gv <- gcGetValues gc
  gcSetValues gc $ gv { foreground = Color 65535 45000 0, lineWidth = 3 }
  mapM_ (drawRoom win gc (segWidth seg) (segHeight seg)) (segRooms seg)
  gcSetValues gc $ gv { foreground = Color 65535 65535 65535 }
  case drag of 
    Just (x,y,w,h) | w>0 && h>0 -> drawRoom win gc (segWidth seg) (segHeight seg) 
                                            (x+1-rmMargin,y+1-rmMargin,w+2*rmMargin,h+2*rmMargin)
    _ -> return ()

-- | Draw a single room outline.                              
drawRoom :: DrawWindow -> GC -> Int -> Int -> (Int,Int,Int,Int) -> IO ()
drawRoom win gc segW segH (x',y',w',h') = do
  let x0 = max 0 (x'*largeW - rmMargin)
      y0 = max 0 (y'*largeH - rmMargin)
      x1 = min (largeW*segW) ((x'+w'-2)*largeW + rmMargin)
      y1 = min (largeH*segH) ((y'+h'-2)*largeH + rmMargin)
  --For some reason drawRectangle doesnt seem to work here. Using drawLines instead.
  drawLines win gc [(x0,y0), (x1-1,y0), (x1-1,y1-1), (x0, y1-1), (x0,y0)]


-- Tile Palette

-- | Helper function to get the size of the tile palette (in tiles)
getTilePaletteSize :: DrawingArea -> Int -> IO (Int, Int)
getTilePaletteSize tileArea nTiles = do
  (widgetWidth, widgetHeight) <- widgetGetSize tileArea
  let xTiles = widgetWidth `div` largeW
  let yTiles = nTiles `divRoundUp` xTiles
  return (xTiles, yTiles)


-- | Draw the tile palette
drawTilePalette :: AppStateRef -> DrawingArea -> Rectangle -> IO ()
drawTilePalette appStateRef tileArea (Rectangle dx dy dw dh) = do
  appState <- readIORef appStateRef
  let gfxLoaderRef = asGfxLoaderRef appState
  let nTiles = length (tiles appState)

  (xTiles, yTiles) <- getTilePaletteSize tileArea nTiles
  widgetSetSizeRequest tileArea (largeW*5) (largeH*yTiles) -- Make sure the window is big enough...

  win <- widgetGetDrawWindow tileArea
  gc <- gcNew win
  pangoContext <- widgetGetPangoContext tileArea
  let doDraw (tile,(x,y)) = drawTile win gc pangoContext gfxLoaderRef tile (x*largeW) (y*largeH)
  mapM_ doDraw (zip (tiles appState) [(x,y) | y<-[0..], x <- [0..(xTiles-1)]])


-- Left/Right Mouse Tiles

-- | Draw the left-mouse and right-mouse "selected tiles"
drawSelectedTile :: AppStateRef -> Bool -> Rectangle -> IO ()
drawSelectedTile appStateRef isLeft _ = do
  appState <- readIORef appStateRef
  let gfxLoaderRef = asGfxLoaderRef appState
      drawingArea = (if isLeft then leftTileDrawingArea else rightTileDrawingArea) (widgets appState)
      tile = (if isLeft then leftMouseTile else rightMouseTile) appState
  win <- widgetGetDrawWindow drawingArea
  gc <- gcNew win
  pangoContext <- widgetGetPangoContext drawingArea
  case tile of 
    Just t -> drawTile win gc pangoContext gfxLoaderRef t 0 0
    Nothing -> return ()
