-- MODULE:
--   GfxLoader
--
-- PURPOSE:
--   Functions to load Knights graphics files and create corresponding
--   GTK Pixbufs.
--
-- AUTHOR:
--   Stephen Thompson <stephen@solarflare.org.uk>
--
-- COPYRIGHT:
--   Copyright (C) Stephen Thompson, 2011.
--
--   This file is part of Knights Map Editor.
--
--   Knights Map Editor is free software: you can redistribute it
--   and/or modify it under the terms of the GNU General Public
--   License as published by the Free Software Foundation, either
--   version 3 of the License, or (at your option) any later version.
--
--   Knights Map Editor is distributed in the hope that it will be
--   useful, but WITHOUT ANY WARRANTY; without even the implied
--   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
--   See the GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with Knights Map Editor. If not, see
--   <http://www.gnu.org/licenses/>.

module GfxLoader
    (GfxLoaderRef,
     newGfxLoader,      -- String -> IO GfxLoaderRef
     loadTileGraphic,   -- GfxLoaderRef -> Tile -> IO Pixbuf   [Int is the tile number.]
    )

where

import qualified MapData -- for Graphic, Tile

import Control.Monad
import Data.IORef
import qualified Data.Map as Map
import Graphics.UI.Gtk
import System.FilePath

type GfxLoaderRef = IORef (Map.Map Int Pixbuf, String)  -- fst = Maps tile numbers to pixbufs; snd = knights_data dir.

newGfxLoader :: String -> IO GfxLoaderRef
newGfxLoader ktsDataDir = newIORef (Map.empty, ktsDataDir)

loadTileGraphic :: GfxLoaderRef -> MapData.Tile -> IO Pixbuf
loadTileGraphic loader tile = do
  let tNumber = MapData.tileNumber tile
      tGfx = MapData.tileGfx tile
      handlePosns = map MapData.gfxHandlePos tGfx
  (ldr, ktsDataDir) <- readIORef loader
  let result = Map.lookup tNumber ldr
  case result of 
    Just pb -> return pb
    Nothing -> do
      -- First time we have seen this tile. Load it from disk
      pixbufs <- mapM (loadGraphicPixbuf ktsDataDir) tGfx
      pixbuf <- compositePixbufs (zipWith (\pb (hx,hy) -> (pb,hx,hy)) pixbufs handlePosns)
      -- save it back into the map
      writeIORef loader (Map.insert tNumber pixbuf ldr, ktsDataDir)
      return pixbuf

loadGraphicPixbuf :: String -> MapData.Graphic -> IO Pixbuf
loadGraphicPixbuf ktsDataDir gfx = do
  let filename = MapData.gfxFilename gfx
      colKey   = MapData.gfxColKey gfx
  pb <- pixbufNewFromFile (ktsDataDir `combine` filename)
  case colKey of 
    Nothing -> return pb
    Just (r,g,b) -> pixbufAddAlpha pb (Just (fromIntegral r, fromIntegral g, fromIntegral b))

compositePixbufs :: [(Pixbuf,Int,Int)] -> IO Pixbuf
compositePixbufs [] = error "no graphic specified for tile"
compositePixbufs [(pb,_,_)] = return pb  -- No compositing needed for just one image
compositePixbufs ((pb,_,_):xs) = foldM composite pb xs

composite :: Pixbuf -> (Pixbuf,Int,Int) -> IO Pixbuf
composite pb1 (pb2,hx,hy) = do   -- We want to draw pb2 over the top of pb1.
  w <- pixbufGetWidth pb2
  h <- pixbufGetHeight pb2
  let x = -hx
      y = (-h+16)+hy
  pixbufComposite pb2 pb1 x y h w (fromIntegral x) (fromIntegral y) 1 1 InterpNearest 255  -- This overwrites pb1, but that is OK
  return pb1
