-- MODULE:
--   KConfig
--
-- PURPOSE:
--   Handles loading of Knights Configuration ("KConfig") files.
--
-- AUTHOR:
--   Stephen Thompson <stephen@solarflare.org.uk>
--
-- COPYRIGHT:
--   Copyright (C) Stephen Thompson, 2011.
--
--   This file is part of Knights Map Editor.
--
--   Knights Map Editor is free software: you can redistribute it
--   and/or modify it under the terms of the GNU General Public
--   License as published by the Free Software Foundation, either
--   version 3 of the License, or (at your option) any later version.
--
--   Knights Map Editor is distributed in the hope that it will be
--   useful, but WITHOUT ANY WARRANTY; without even the implied
--   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
--   See the GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with Knights Map Editor. If not, see
--   <http://www.gnu.org/licenses/>.


module KConfig 
    (KValue,          -- abstract type. instance of Show
     kInt,            -- KValue -> Maybe Int
     kFloat,          -- KValue -> Maybe Float
     kString,         -- KValue -> Maybe String
     kTableLookup,    -- String -> KValue -> Maybe KValue
     kList,           -- KValue -> Maybe [KValue]
     kDirective,      -- KValue -> Maybe (String, KValue)
     kSymbol,         -- KValue -> Maybe String    -- Get symbol name w/o dereferencing

     loadKConfig,     -- String -> (String -> IO String) -> IO (Either String KValue)   -- returns a table
     loadKConfigFile, -- String -> IO (Either String KValue)
     parseKValue,     -- String -> Either String KValue   -- the string is an error msg

     makeKInt,        -- Int -> KValue
     makeKString,     -- String -> KValue
     makeKList,       -- [KValue] -> KValue
     makeKTable       -- Map String KValue -> KValue
    )

where

import Control.Monad.Loops
import Data.List
import qualified Data.Map as Map
import Data.Maybe
import GoldParser
import KConfigCgt
import Numeric
import qualified Scripting.Lua as Lua
import System.IO


-- TYPE DEFINITIONS

type KSymbolTable = Map.Map String KValue

-- Note that the individual KValue constructors are hidden (KValue is an abstract type).
-- Use the functions below to read KValues.
data KValue = KInt Int
            | KFloat Float
            | KString String
            | KRandomInt                         -- TODO We don't load random int info yet
            | KTable KSymbolTable
            | KList [(KValue,Int)]               -- list of values / repeat counts.
            | KDirective String KValue           -- name of directive + list of arguments.
            | KSymbol String (Maybe KValue)      -- Symbol name, plus the referenced value if there is one (latter evaluated lazily of course). 
            | KBinExp Int KValue KValue          -- reduction number, then lhs, then rhs.
            | KUnaryMinus KValue
              deriving (Eq, Ord)


-- LUA FUNCTIONS

-- throw error if result is nonzero
luaErrChk :: Lua.LuaState -> Int -> String -> IO ()
luaErrChk lua result filename = do
  if result /= 0 then do
    msg <- Lua.tostring lua (-1)
    error $ "Lua Error: could not load " ++ filename ++ ". Error msg = " ++ msg
   else
    return ()

-- Auxiliary function to load and run a Lua chunk
runChunk :: Lua.LuaState -> String -> String -> IO ()
runChunk lua text filename = do
  result <- Lua.loadstring lua text filename
  luaErrChk lua result filename
  -- the chunk is now at top of stack
  result2 <- Lua.call lua 0 0
  -- the chunk has now been popped.
  luaErrChk lua result2 filename

-- Load the Lua state
openLua :: (String -> IO String) -> IO Lua.LuaState
openLua includeLoader = do
  editor <- includeLoader "editor.lua"
  luaTest <- includeLoader "lua_test.lua"
  lua <- Lua.newstate
  Lua.openlibs lua
  runChunk lua editor "editor.lua"
  runChunk lua luaTest "lua_test.lua"
  return lua

-- Parse out a single KValue from the table at top of stack
idxKValue :: Lua.LuaState -> Int -> IO KValue
idxKValue lua idx = do
  Lua.rawgeti lua (-1) idx
  -- value is now on top of stack
  -- NOTE: only supports "graphics" for now, which are string in lead position, followed by ints.
  kv <- if (idx==1) then do
          s <- Lua.tostring lua (-1)
          return $ KString s
         else do
          i <- Lua.tointeger lua (-1)
          return $ KInt (fromIntegral i)
  Lua.pop lua 1
  return kv

-- Parse out the contents of the table at top of stack.
-- Return a corresponding KValue.
-- (Only KLists supported currently)  
readKValue :: Lua.LuaState -> IO KValue
readKValue lua = do
  len <- Lua.objlen lua (-1)
  lst <- mapM (idxKValue lua) [1..len]
  return $ KList $ (map (\x -> (x,1))) lst
  
-- Determine if the value at top of stack is a "KTable"
isKTable :: Lua.LuaState -> IO Bool
isKTable lua = do
  isTbl <- Lua.istable lua (-1)    -- [table]
  if isTbl then do
    Lua.pushstring lua "kvalue_tag"   -- [table "kvalue_tag"]
    Lua.gettable lua (-2)             -- [table (tag value)]
    isnil <- Lua.isnil lua (-1)
    Lua.pop lua 1     -- [table]
    return (not isnil)    -- If it wasn't nil then this is a KTable
   else
    return False   -- This isn't a table so can't be a KTable.

-- Examine key, value at top of stack
-- If a valid KValue definition then return it,
-- otherwise return Nothing.
processLuaDef :: Lua.LuaState -> IO (Maybe (String, KValue))
processLuaDef lua = do
  isKeyString <- Lua.isstring lua (-2)
  keyString <- if isKeyString then (Lua.tostring lua (-2)) else (return "")
  isKValue <- isKTable lua
  result <- if isKValue then do
              kv <- readKValue lua
              return $ Just (keyString, kv)
            else
              return Nothing
  Lua.pop lua 1   -- pop the value (not the key)
  return result

-- Master function to return a KSymbolTable from the lua state
loadLua :: (String -> IO String) -> IO KSymbolTable
loadLua includeLoader = do
  -- load the lua files
  lua <- openLua includeLoader
  
  -- iterate through the global table, find all KValues (tables tagged with "kvalue_tag")
  Lua.pushnil lua
  results' <- whileM (Lua.next lua Lua.globalsindex) (processLuaDef lua)
  let results = catMaybes results'  :: [(String, KValue)]

  -- return result
  return $ Map.fromList results



-- KCONFIG FUNCTIONS

-- helper for expression evaluation
doExp op (Just x) (Just y) = Just (op x y)
doExp op _ _ = Nothing

-- Functions to read KValues and convert them into usable types.
kInt :: KValue -> Maybe Int
kInt (KSymbol _ (Just val)) = kInt val
kInt (KBinExp 1 x y) = doExp (+) (kInt x) (kInt y)
kInt (KBinExp 2 x y) = doExp (-) (kInt x) (kInt y)
kInt (KBinExp 5 x y) = doExp (*) (kInt x) (kInt y)
kInt (KBinExp 6 x y) = doExp div (kInt x) (kInt y)
kInt (KUnaryMinus x) = doExp (-) (Just 0) (kInt x)
kInt (KInt i) = Just i
kInt _ = Nothing

kFloat :: KValue -> Maybe Float
kFloat (KSymbol _ (Just val)) = kFloat val
kFloat (KBinExp 1 x y) = doExp (+) (kFloat x) (kFloat y)
kFloat (KBinExp 2 x y) = doExp (-) (kFloat x) (kFloat y)
kFloat (KBinExp 5 x y) = doExp (*) (kFloat x) (kFloat y)
kFloat (KBinExp 6 x y) = doExp (/) (kFloat x) (kFloat y)
kFloat (KUnaryMinus x) = doExp (-) (Just 0) (kFloat x)
kFloat (KInt x) = Just (fromIntegral x)
kFloat (KFloat x) = Just x
kFloat _ = Nothing

kString :: KValue -> Maybe String
kString (KSymbol _ (Just val)) = kString val
kString (KString s) = Just s
kString _ = Nothing

kTableLookup :: String -> KValue -> Maybe KValue
kTableLookup key (KSymbol _ (Just val)) = kTableLookup key val
kTableLookup key (KBinExp 3 x y) = case kTableLookup key y of
                                     Just result -> Just result
                                     Nothing -> kTableLookup key x
kTableLookup key (KTable m) = Map.lookup key m
kTableLookup key _ = Nothing

kList :: KValue -> Maybe [KValue]   -- expands out repeat counts. 
kList (KSymbol _ (Just val)) = kList val
kList (KList lst) = Just $ concatMap (\(v,rc)->replicate rc v) lst
kList _ = Nothing

kDirective :: KValue -> Maybe (String, KValue)
kDirective (KSymbol _ (Just val)) = kDirective val
kDirective (KDirective n args) = Just (n,args)
kDirective _ = Nothing

kSymbol :: KValue -> Maybe String
kSymbol (KSymbol x _) = Just x
kSymbol _ = Nothing


-- Load the cgt...
kConfigLoadedCgt = loadCgt kConfigCgt 

-- on success, returns symbol table from this file, and list of included files.
doLoadKConfig :: [Char] -> KSymbolTable -> KSymbolTable -> Either String (KSymbolTable, [String])
doLoadKConfig stream inputSymTbl finalSymTbl = 
    let initialState = initialParseState kConfigLoadedCgt stream
        initialStack = [KTable inputSymTbl]
    in case doParse initialState (initialStack,[]) finalSymTbl of
         Left e -> Left e
         Right (finalStack, includedFiles) -> 
             let (KTable finalSymTbl') = head finalStack
             in Right (finalSymTbl', includedFiles)

doParse :: ParseState -> ([KValue],[String]) -> KSymbolTable -> Either String ([KValue], [String])
doParse state stackAndFiles finalSymTbl = 
    let (msg, newState) = parse state
    in case msg of
         Shift sym str -> doParse newState (shift  sym str stackAndFiles finalSymTbl) finalSymTbl
         Reduce rule -> 
             case (reduce rule stackAndFiles) of
               Left e -> Left e
               Right r -> doParse newState r finalSymTbl
         Accept        -> Right stackAndFiles
         SyntaxError   -> Left "Syntax error"
         LexicalError  -> Left "Lexical error"
         CommentError  -> Left "Comment error"


shift :: Symbol -> String -> ([KValue],[String]) -> KSymbolTable -> ([KValue],[String])
shift sym tok (stack,files) finalSymTbl = 
    case (shiftResult sym tok finalSymTbl) of 
      Just val -> (val : stack, files)
      Nothing  -> (stack, files)

shiftResult :: Symbol -> String -> KSymbolTable -> Maybe KValue
shiftResult (Sym 18) _   _= Just $ KRandomInt  -- dice
shiftResult (Sym 19) str _= Just $ KFloat (read str)
shiftResult (Sym 20) str _= Just $ KInt ((fst.head.readHex) (tail str))  -- the "tail" strips off the leading $
shiftResult (Sym 22) str _= Just $ KInt (read str)
shiftResult (Sym 23) str _= Just $ KString ((tail.init) str)  -- strip off quotes at beginning and end
shiftResult (Sym 24) str finalSymTbl = Just $ KSymbol str (Map.lookup str finalSymTbl)
shiftResult _        _   _= Nothing


-- Reduce takes rule number and stack/files, and returns new stack/files.
reduce :: Int -> ([KValue],[String]) -> Either String ([KValue],[String])
reduce 30 ((KString filename) : stk, files) = Right (stk, filename:files) -- Include
reduce r (stack,files) =           -- Anything else (doesn't touch includedFiles list).
    case reduceResult r stack of
      Left e -> Left e
      Right x -> Right (x,files)

-- Assignment
reduceResult :: Int -> [KValue] -> Either String [KValue]
reduceResult 0 (target : (KSymbol symName _) : (KTable table) : stk) = 
    case (Map.lookup symName table) of
      Just _ -> Left "multiple definition"
      Nothing ->
          let newTable = Map.insert symName target table
          in Right $ (KTable newTable) : stk

-- Binary operations
reduceResult r (rhs : lhs : stk) 
    | r==1 || r==2 || r==3 || r==5 || r==6 =
        Right $ (KBinExp r lhs rhs) : stk

-- Unary minus
reduceResult 17 (val : stk) = Right $ (KUnaryMinus val) : stk

-- New table
reduceResult 21 stk = Right $ (KTable Map.empty) : stk

-- New list
reduceResult 24 stk = Right $ (KList []) : stk

-- Add list (31 = single item, 32 = multiple items)
-- Note stupid way of doing this by appending to back (could change the grammar to fix this I suppose).
reduceResult 31 stk = reduceResult 32 (KInt 1 : stk)
reduceResult 32 ((KInt count) : val : (KList lst) : stk)
    = Right $ (KList (lst++[(val,count)])) : stk

-- Directive
reduceResult 33 (args : (KSymbol directiveName _) : stk)
    = Right $ (KDirective directiveName args) : stk

-- Default: no-operation.
reduceResult _ stk = Right stk


-- Load a "KConfig" given a way of loading the individual include-files. 
-- The filename to use for the root file must be given. 
loadKConfig :: String -> (String -> IO [Char]) -> IO (Either String KValue)
loadKConfig rootFile includeLoader = do
  luaSymbols <- loadLua includeLoader
  result <- fixIO (\r -> loadAlgo luaSymbols (getFinalSymTbl r) [rootFile] [] includeLoader)
  return $ case result of
             Left x -> Left x
             Right y -> Right (KTable y)

getFinalSymTbl :: Either String KSymbolTable -> KSymbolTable
getFinalSymTbl (Left e) = Map.empty
getFinalSymTbl (Right x) = x  

loadAlgo :: KSymbolTable -> KSymbolTable -> [String] -> [String] -> (String -> IO [Char]) -> IO (Either String KSymbolTable)
loadAlgo symTblSoFar finalSymTbl filesToLoad alreadyIncludedFiles includeLoader = do
  case filesToLoad of 
    [] -> return (Right symTblSoFar) -- Done
    (f:fs) -> 
        case find (==f) alreadyIncludedFiles of
          Just _  ->
            -- This file has already been included; move on to the next one
            loadAlgo symTblSoFar finalSymTbl fs alreadyIncludedFiles includeLoader
          Nothing -> do
            -- Go ahead and load this include-file.
            newStream <- includeLoader f
            case doLoadKConfig newStream symTblSoFar finalSymTbl of
              Left e -> return $ Left e
              Right (newSymTbl, newlyIncludedFiles) ->
                  loadAlgo newSymTbl finalSymTbl (newlyIncludedFiles++fs) (f:alreadyIncludedFiles) includeLoader

-- Load a KConfig from a file in the current directory
loadKConfigFile :: String -> IO (Either String KValue)
loadKConfigFile filename = 
  let loader :: String -> IO [Char]
      loader name = do
        f <- openFile name ReadMode
        hGetContents f
  in loadKConfig filename loader



-- helper function for showing expressions

instance Show KValue where
    show (KInt i) = show i
    show (KFloat x) = show x
    show (KString str) = show str
    show (KRandomInt) = error "don't know how to show random ints"  -- TODO. We don't load random ints currently.
    show (KTable symTbl) = 
        "{" ++ (intercalate ", " $ Map.elems $ Map.mapWithKey showval symTbl) ++ "}"
            where showval k v = k ++ " = " ++ (show v)
    show (KList lst) = 
        "[" ++ (intercalate ", " $ map show (fromJust $ kList (KList lst))) ++ "]"
    show (KDirective s (KList v)) = s ++ "(" ++ (showargs v) ++ ")"
        where showargs args = intercalate ", " $ map showarg args
              showarg (k,1) = (show k)
              showarg (k,n) = (show k) ++ ":" ++ (show n)
    show (KSymbol s _) = s
    show (KBinExp n x y) = formatBinOp n x y
    show (KUnaryMinus x) = "-" ++ (show x)

formatBinOp :: Int -> KValue -> KValue -> String
formatBinOp n left right = 
    let leftStr = show left
        rightStr = show right
        leftPri = priority left (head leftStr == '(') False
        rightPri = priority right (head rightStr == '(') True
        opPri = opToPri n False
        bracketLeft = leftPri < opPri
        bracketRight = rightPri < opPri
        leftStr' = if bracketLeft then "(" ++ leftStr ++ ")" else leftStr
        rightStr' = if bracketRight then "(" ++ rightStr ++ ")" else rightStr
        --debug = "[" ++ show leftPri ++ "/" ++ show opPri ++ "/" ++ show rightPri ++ "]"
        debug = ""
    in leftStr' ++ opToStr n ++ debug ++ rightStr'
       

priority :: KValue 
         -> Bool   -- ^ bracketed?
         -> Bool   -- ^ right?
         -> Int
priority _ True _ = 10
priority (KBinExp n _ _) False right = opToPri n right
priority _ False _ = 10

opToPri :: Int    -- ^ operator number
        -> Bool   -- ^ on right?
        -> Int
opToPri 1 _     = 1   -- +
opToPri 2 False = 1   -- -
opToPri 2 True  = 0   -- - at right
opToPri 5 _     = 3   -- *
opToPri 6 False = 3   -- /
opToPri 6 True  = 2   -- / at right
opToPri 3 _     = 5   -- &

opToStr :: Int -> String
opToStr 1 = "+"
opToStr 2 = "-"
opToStr 3 = "&"
opToStr 5 = "*"
opToStr 6 = "/"



parseKValue :: String -> Either String KValue
parseKValue string =
    let result = doLoadKConfig ("__test_sym__ = " ++ string) (Map.empty) (getFinalSymTbl2 result) in
    case result of 
      Left e -> Left e
      Right (finalSymTbl, []) ->
          if Map.size finalSymTbl /= 1 then Left "Syntax error"
          else case Map.lookup "__test_sym__" finalSymTbl of
                 Nothing -> Left "problem in parseKValue"
                 Just x -> Right x


getFinalSymTbl2 :: Either String (KSymbolTable, a) -> KSymbolTable
getFinalSymTbl2 (Left _) = Map.empty
getFinalSymTbl2 (Right (x,_)) = x



makeKInt = KInt
makeKString = KString
makeKTable = KTable

-- TODO could perhaps 'compress' lists i.e. use repeat counts. currently the repeat count is always set to 1.
makeKList :: [KValue] -> KValue
makeKList lst = KList (zip lst (repeat 1))


