-- MODULE:
--   Main
--
-- PURPOSE:
--   Main program for Knights Map Editor.
--
-- AUTHOR:
--   Stephen Thompson <stephen@solarflare.org.uk>
--
-- COPYRIGHT:
--   Copyright (C) Stephen Thompson, 2011.
--
--   This file is part of Knights Map Editor.
--
--   Knights Map Editor is free software: you can redistribute it
--   and/or modify it under the terms of the GNU General Public
--   License as published by the Free Software Foundation, either
--   version 3 of the License, or (at your option) any later version.
--
--   Knights Map Editor is distributed in the hope that it will be
--   useful, but WITHOUT ANY WARRANTY; without even the implied
--   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
--   See the GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with Knights Map Editor. If not, see
--   <http://www.gnu.org/licenses/>.

module Main where 

import AppState
import KConfig (KValue, parseKValue)
import ReadXml

import Control.Monad (when)
import Data.List
import Data.Maybe
import Graphics.UI.Gtk hiding ( eventButton, eventClick, eventIsHint, eventKeyVal, eventModifier, eventSent, eventTime )
import Graphics.UI.Gtk.Builder
import Graphics.UI.Gtk.Gdk.Events
import System.Directory
import System.FilePath
import System.IO
import System.IO.Error


-- File names for the config files
knightsEditorAppName = "knights_editor"   -- Parameter to System.Directory.getAppUserDataDirectory
knightsEditorConfigFileName = "knights_editor_config.txt"


-- Event Handlers

-- | Mouse motion within the map area
mapMouseMotion :: AppStateRef -> Event -> IO Bool
mapMouseMotion appStateRef ev = do
  -- Get mouse coordinates
  let x = floor (eventX ev)
      y = floor (eventY ev)
  -- Calculate tile coordinates that mouse is now over
  let (newTX, newTY) = mouseToTileCoords x y
  -- Update the mouse hover position
  setMouseHover appStateRef $ Just (newTX, newTY)
  -- Update the map (paint new tile) if necessary
  em <- getEditMode appStateRef
  when (em == PaintMode) $ do
      mb <- getMouseButton appStateRef
      case mb of 
        Just LeftButton -> doSetTile appStateRef getLeftMouseTile newTX newTY
        Just RightButton -> doSetTile appStateRef getRightMouseTile newTX newTY
        _ -> return ()
  return True

doSetTile appStateRef getTile newTX newTY = do
  t <- getTile appStateRef
  case t of 
    Just tile -> writeTileToCurrSeg appStateRef tile newTX newTY
    Nothing -> return ()


-- | Called when mouse exits the map area
mapMouseExit :: AppStateRef -> Event -> IO Bool
mapMouseExit appStateRef _ = do
  cancelMouseDrag appStateRef
  mh <- getMouseHover appStateRef
  case mh of 
    Nothing -> return True -- Nothing to do
    Just (x,y) -> do
      -- Clear the 'hover' state
      setMouseHover appStateRef Nothing
      setMouseButton appStateRef Nothing
      return True

-- | Mouse click within the map area
mapMouseDown :: Dialog -> Entry -> AppStateRef -> Event -> IO Bool
mapMouseDown switchDialog switchEntry appStateRef ev = do
  -- Mouse coordinates
  let mx = floor (eventX ev)
      my = floor (eventY ev)
  -- Tile coordinates
  let (tx, ty) = mouseToTileCoords mx my
  editMode <- getEditMode appStateRef
  let mouseButton = eventButton ev
      ctrlPressed = Control `elem` (eventModifier ev)
  case (eventClick ev) of
    SingleClick | editMode == PaintMode && ctrlPressed -> do
      -- Click (either left or right) with alt pressed = Select tile
      selectTileFromMap (mouseButton == LeftButton) appStateRef tx ty
      return True
    SingleClick | editMode == SwitchesMode && mouseButton == RightButton -> do
      -- Right-click in switches mode = Delete switch
      rmSwitch appStateRef tx ty
      return True
    SingleClick | editMode == RoomsMode && mouseButton == RightButton -> do
      -- Right-click in rooms mode = Delete room
      deleteRoomAt appStateRef tx ty
      return True
    SingleClick -> do 
      -- First check if a different button is already held down
      justOldButton <- getMouseButton appStateRef
      case justOldButton of 
        Just oldButton | mouseButton /= oldButton -> do
          -- OK, in this case we just cancel everything (as if the button had been released)
          mapMouseUp appStateRef ev
        _ -> do
          -- Update appState with the new mouse button status, then call through to the "mouse motion" event
          setMouseButton appStateRef (Just mouseButton)
          doMotionEvent appStateRef ev
          when (mouseButton == LeftButton) (startMouseDrag appStateRef tx ty)
          return True
    DoubleClick | (editMode == SwitchesMode && mouseButton == LeftButton) -> do
      -- Double click in switches mode -- brings up switch dialog
      s <- isSegSelected appStateRef
      when s $ do
        currSwitch <- getSwitch appStateRef tx ty
        newSwitch <- runSwitchDialog switchDialog switchEntry currSwitch
        case newSwitch of
          Nothing -> return ()
          Just ns -> setSwitch appStateRef tx ty ns
      return True
    _ -> return False


-- | Display the switch dialog.
runSwitchDialog :: Dialog
                -> Entry
                -> Maybe KValue   -- ^ Initial value to display in the dialog
                -> IO (Maybe KValue)   -- ^ Returned value, or Nothing if cancelled
runSwitchDialog switchDialog switchEntry input = do
  let initialTxt = case input of
                     Nothing -> ""
                     Just k -> show k
  entrySetText switchEntry initialTxt
  resultCode <- dialogRun switchDialog
  widgetHide switchDialog
  case resultCode of
    ResponseOk -> do
      newTxt <- entryGetText switchEntry
      -- TODO handle parse errors: should probably display some kind of error, maybe also disallow dialog exit if we can.
      case parseKValue newTxt of
        Right kv -> return $ Just kv
        _ -> return Nothing
    _ -> return Nothing


-- | Mouse button released within map area
mapMouseUp :: AppStateRef -> Event -> IO Bool
mapMouseUp appStateRef ev = do
  -- After a button is released, reset everything.
  oldMB <- getMouseButton appStateRef
  editMode <- getEditMode appStateRef
  setMouseButton appStateRef Nothing
  when (oldMB == Just LeftButton && editMode == RoomsMode) (setRoomFromDrag appStateRef)
  cancelMouseDrag appStateRef
  doMotionEvent appStateRef ev
  return True

-- | The motion event following a mouse event.
doMotionEvent :: AppStateRef -> Event -> IO ()
doMotionEvent appStateRef ev = do
  mapMouseMotion appStateRef (Motion { eventSent = undefined, eventTime = undefined, 
                                       eventX = eventX ev, eventY = eventY ev,
                                       eventModifier = undefined, eventIsHint = undefined,
                                       eventXRoot = undefined, eventYRoot = undefined })
  return ()



-- | Handle clicks in the segments area.
segsMouseDown :: AppStateRef -> Event -> IO Bool
segsMouseDown appStateRef ev = do
  case (eventClick ev) of
    DoubleClick -> do
      -- Mouse coordinates
      let x = floor (eventX ev)
          y = floor (eventY ev)

      -- Convert to a segment name (if any)
      segClicked <- mouseToSegmentName appStateRef x y

      -- Do the update
      case segClicked of 
        Just segName -> setCurrentSegment appStateRef segName
        Nothing -> return ()
      return True
    _ -> return False

-- | Mouse wheel moved within segments area
segsScroll :: AppStateRef -> VScrollbar -> Event -> IO Bool
segsScroll appStateRef scrollbar ev = do
  let mult = case eventDirection ev of
               ScrollUp -> -1.5
               ScrollDown -> 1.5
               _ -> 0
  adj <- rangeGetAdjustment scrollbar
  currVal <- adjustmentGetValue adj
  inc <- adjustmentGetStepIncrement adj
  lower <- adjustmentGetLower adj
  upper <- adjustmentGetUpper adj
  pageSize <- adjustmentGetPageSize adj
  let newVal = max lower (min (upper - pageSize) (currVal + inc * mult))
  adjustmentSetValue adj newVal
  return True

-- | A mouse click in the tile area - Selects a new tile
tilePaletteMouseDown :: AppStateRef -> DrawingArea -> Event -> IO Bool
tilePaletteMouseDown appStateRef tileArea ev = do
  case (eventClick ev) of
    SingleClick -> do
      -- Mouse coordinates
      let x = floor (eventX ev)
          y = floor (eventY ev)

      -- Convert to a tile number
      tileNum <- mouseToTileNum tileArea x y

      -- Do the update (this does nothing if tileNum is out of range).
      setSelectedTile (eventButton ev == LeftButton) appStateRef tileNum

    _ -> return ()
  return True


-- | Resets the edit mode when one of the toggle buttons is pressed
toggleButtonPressed :: ToggleButtonClass tb => AppStateRef -> tb -> EditMode -> IO ()
toggleButtonPressed appStateRef button newMode = do
  t <- toggleButtonGetActive button
  when t $ setEditMode appStateRef newMode


-- | Prompt user to save their work if the file is changed.
promptSaveIfNeeded :: AppStateRef -> (AppStateRef -> IO ()) -> IO ()
promptSaveIfNeeded appStateRef func = do
  unsaved <- isUnsaved appStateRef
  if unsaved then do
        dialog <- messageDialogNew Nothing [] MessageQuestion ButtonsNone "Save file before closing?"
        dialogAddButton dialog stockYes ResponseYes
        dialogAddButton dialog stockNo ResponseNo
        dialogAddButton dialog stockCancel ResponseCancel
        result <- dialogRun dialog
        widgetDestroy dialog
        case result of
          ResponseYes -> do
                      ok <- saveKtsFile appStateRef
                      if ok then func appStateRef else return ()
          ResponseNo -> func appStateRef
          ResponseCancel -> return ()
   else func appStateRef    -- File is saved already, so proceed without asking for confirmation.

-- | Open a new file
openKtsFile :: AppStateRef -> IO ()
openKtsFile appStateRef = do
  dialog <- fileChooserDialogNew (Just "Open File") Nothing FileChooserActionOpen 
               [("Open", ResponseAccept), ("Cancel", ResponseReject)]
  resultCode <- dialogRun dialog
  filename <- fileChooserGetFilename dialog
  when (resultCode == ResponseAccept && isJust filename) 
     (loadRoomsFile appStateRef (fromJust filename))
  widgetDestroy dialog

-- | Save the current file
-- Returns True if actually saved, False if user cancelled (latter can only happen
-- if current file is unnamed, so a save as dialog comes up).
saveKtsFile :: AppStateRef -> IO Bool
saveKtsFile asr = do
  isNewFile <- isFileNameEmpty asr
  if isNewFile 
   then saveKtsFileAs asr 
   else (saveRoomsFile asr >> return True)

-- | Save under a new name
-- Returns True if actually saved, False if user cancelled
saveKtsFileAs :: AppStateRef -> IO Bool
saveKtsFileAs appStateRef = do
  dialog <- fileChooserDialogNew (Just "Save File") Nothing FileChooserActionSave
               [("Save", ResponseAccept), ("Cancel", ResponseReject)]
  resultCode <- dialogRun dialog
  filename <- fileChooserGetFilename dialog
  let ok = resultCode == ResponseAccept && isJust filename
  when ok (saveRoomsFileAs appStateRef (fromJust filename))
  widgetDestroy dialog
  return ok

promptSaveAndExit :: AppStateRef -> IO ()
promptSaveAndExit appStateRef = promptSaveIfNeeded appStateRef (\_ -> closeDownProgram appStateRef)

closeDownProgram :: AppStateRef -> IO ()
closeDownProgram asr = cleanUpTempFiles asr True >> mainQuit

-- | Show the dialog to edit the config (knights_data path etc).
-- This function will write a new config file if necessary.
runConfigDialog :: Dialog -> Entry -> Entry -> AppStateRef -> IO ()
runConfigDialog configDialog ktsDataEntry ktsExeEntry appStateRef = do
  (oldKtsDataDir, oldKtsExePath) <- getKnightsPaths appStateRef
  entrySetText ktsDataEntry oldKtsDataDir
  entrySetText ktsExeEntry oldKtsExePath
  resultCode <- dialogRun configDialog
  widgetHide configDialog
  when (resultCode == ResponseOk) $ do
    ktsDataDir <- entryGetText ktsDataEntry
    ktsExePath <- entryGetText ktsExeEntry
    setKnightsPaths appStateRef ktsDataDir ktsExePath
    writeConfigFile ktsDataDir ktsExePath
  return ()

browseForFile :: Entry -> FileChooserAction -> IO ()
browseForFile entry fcAction = do
  dialog <- fileChooserDialogNew Nothing Nothing fcAction [("Select", ResponseAccept), ("Cancel", ResponseReject)]
  resultCode <- dialogRun dialog
  filename <- fileChooserGetFilename dialog
  let ok = resultCode == ResponseAccept && isJust filename
  when ok (entrySetText entry (fromJust filename))
  widgetDestroy dialog

-- | Hide the "tile brush" when ctrl key is pressed.
mainWindowKeyEvent :: AppStateRef -> Event -> IO Bool
mainWindowKeyEvent appStateRef ev = do
  let k = eventKeyVal ev
  if k == (keyFromName "Control_L") || k == (keyFromName "a") then do
    setShowTileBrush appStateRef (eventRelease ev)
    return True
   else return False


-- | Tries to read the config file. Returns the knights_data and knights.exe locations, or "" if there was an error
-- reading the config file.
readConfigFile :: IO (String, String)
readConfigFile = do
  result <- try $ do
    myAppDir <- getAppUserDataDirectory knightsEditorAppName
    createDirectoryIfMissing False myAppDir
    withFile (myAppDir `combine` knightsEditorConfigFileName) ReadMode $ \h -> do
      ktsDataDir <- hGetLine h
      ktsExeLocn <- hGetLine h
      return (ktsDataDir, ktsExeLocn)
  case result of
    Left _ -> return ("", "")
    Right a -> return a

-- | Write the knights config file to disk
-- Write errors are silently ignored.
writeConfigFile :: String -> String -> IO ()
writeConfigFile dataDir exePath = do
  try $ do
    myAppDir <- getAppUserDataDirectory knightsEditorAppName
    createDirectoryIfMissing False myAppDir
    withFile (myAppDir `combine` knightsEditorConfigFileName) WriteMode $ \h -> do
      hPutStrLn h dataDir
      hPutStrLn h exePath
  return ()
    

-- | "Main routine" for the application. Starts the Main Window.
main :: IO ()
main = do
-- Initialize GTK
  initGUI

-- Load xml
  xml <- readXml "knightseditor.glade"

-- Get widgets of interest
  mainWindow <- builderGetObject xml castToWindow "mainWindow"
  segmentsDrawingArea <- builderGetObject xml castToDrawingArea "segmentsDrawingArea"
  segmentsScrollbar <- builderGetObject xml castToVScrollbar "segmentsScrollBar"
  mapDrawingArea <- builderGetObject xml castToDrawingArea "mapDrawingArea"
  leftSelectedTile <- builderGetObject xml castToDrawingArea "leftSelectedTile"
  rightSelectedTile <- builderGetObject xml castToDrawingArea "rightSelectedTile"
  tileDrawingArea <- builderGetObject xml castToDrawingArea "tileDrawingArea"
  switchesTreeView <- builderGetObject xml castToTreeView "switchesTreeView"
  paintButton <- builderGetObject xml castToToggleButton "paintButton"
  switchesButton <- builderGetObject xml castToToggleButton "switchesButton"
  roomsButton <- builderGetObject xml castToToggleButton "roomsButton"
  switchDialog <- builderGetObject xml castToDialog "switchDialog"
  switchEntry <- builderGetObject xml castToEntry "switchEntry"
  nameEntry <- builderGetObject xml castToEntry "nameEntry"
  categoryEntry <- builderGetObject xml castToEntry "categoryEntry"
  configDialog <- builderGetObject xml castToDialog "configDialog"
  knightsDataEntry <- builderGetObject xml castToEntry "knightsDataEntry"
  knightsDataBrowse <- builderGetObject xml castToButton "knightsDataBrowse"
  knightsExeEntry <- builderGetObject xml castToEntry "knightsExeEntry"
  knightsExeBrowse <- builderGetObject xml castToButton "knightsExeBrowse"
  
-- Get menu items of interest
  newItem <- builderGetObject xml castToMenuItem "newItem"
  openItem <- builderGetObject xml castToMenuItem "openItem"
  saveItem <- builderGetObject xml castToMenuItem "saveItem"
  saveAsItem <- builderGetObject xml castToMenuItem "saveAsItem"
  configItem <- builderGetObject xml castToMenuItem "configItem"
  quitItem <- builderGetObject xml castToMenuItem "quitItem"
  addSegmentItem <- builderGetObject xml castToMenuItem "addSegmentItem"
  deleteSegmentItem <- builderGetObject xml castToMenuItem "deleteSegmentItem"
  testSegmentItem <- builderGetObject xml castToMenuItem "testSegmentItem"
  undoItem <- builderGetObject xml castToMenuItem "undoItem"
  redoItem <- builderGetObject xml castToMenuItem "redoItem"

-- Create the AppState
  let widgets = MainWindowWidgets segmentsDrawingArea segmentsScrollbar mapDrawingArea leftSelectedTile
                                  rightSelectedTile tileDrawingArea nameEntry categoryEntry
  appStateRef <- newAppState widgets switchesTreeView
  
-- Load the knights data dir from the config, if it exists
  (ktsDataDir, ktsExeLocation) <- readConfigFile
  setKnightsPaths appStateRef ktsDataDir ktsExeLocation
  
-- Set up events

  -- Main Window
  mainWindow `onDelete` (\_ -> promptSaveAndExit appStateRef >> return True)

  -- Segments drawing area
  segmentsDrawingArea `onExposeRect` (drawThumbs appStateRef)
  segmentsDrawingArea `onButtonPress` (segsMouseDown appStateRef)
  segmentsDrawingArea `onScroll` (segsScroll appStateRef segmentsScrollbar)
  segmentsScrollbar `onRangeValueChanged` (widgetQueueDraw segmentsDrawingArea)

  -- Map drawing area
  -- NOTE: for onMotionNotify, "True" means mouse "hints" only (more efficient apparently)
  mapDrawingArea `onExposeRect` (drawMap appStateRef)
  onMotionNotify mapDrawingArea True (mapMouseMotion appStateRef)
  mapDrawingArea `onLeaveNotify` (mapMouseExit appStateRef)
  mapDrawingArea `onButtonPress` (mapMouseDown switchDialog switchEntry appStateRef)
  mapDrawingArea `onButtonRelease` (mapMouseUp appStateRef)

  -- Selected tiles drawing areas
  leftSelectedTile `onExposeRect` (drawSelectedTile appStateRef True)
  rightSelectedTile `onExposeRect` (drawSelectedTile appStateRef False)

  -- Tile drawing area (tile palette)
  tileDrawingArea `onButtonPress` (tilePaletteMouseDown appStateRef tileDrawingArea)
  tileDrawingArea `onExposeRect` (drawTilePalette appStateRef tileDrawingArea)

  -- other buttons etc
  paintButton `onToggled` (toggleButtonPressed appStateRef paintButton PaintMode)
  switchesButton `onToggled` (toggleButtonPressed appStateRef switchesButton SwitchesMode)
  roomsButton `onToggled` (toggleButtonPressed appStateRef roomsButton RoomsMode)

  -- Segment Properties - "fixed" when user presses return.
  nameEntry `onEntryActivate` (updateSegName appStateRef)
  categoryEntry `onEntryActivate` (updateSegCategory appStateRef)

  -- Menu Items
  newItem `onActivateLeaf` (promptSaveIfNeeded appStateRef resetAppState)
  openItem `onActivateLeaf` (promptSaveIfNeeded appStateRef openKtsFile)
  saveItem `onActivateLeaf` (saveKtsFile appStateRef >> return ())
  saveAsItem `onActivateLeaf` (saveKtsFileAs appStateRef >> return ())
  configItem `onActivateLeaf` (runConfigDialog configDialog knightsDataEntry knightsExeEntry appStateRef)
  quitItem `onActivateLeaf` (promptSaveAndExit appStateRef)
  addSegmentItem `onActivateLeaf` (addNewSegment appStateRef)
  deleteSegmentItem `onActivateLeaf` (deleteCurrSegment appStateRef)
  testSegmentItem `onActivateLeaf` (testCurrentSegment appStateRef)
  undoItem `onActivateLeaf` (undo appStateRef)
  redoItem `onActivateLeaf` (redo appStateRef)

  -- Config dialog events
  knightsDataBrowse `onClicked` (browseForFile knightsDataEntry FileChooserActionSelectFolder)
  knightsExeBrowse `onClicked` (browseForFile knightsExeEntry FileChooserActionOpen)

  -- Key events
  mainWindow `onKeyPress` (mainWindowKeyEvent appStateRef)
  mainWindow `onKeyRelease` (mainWindowKeyEvent appStateRef)

  -- Periodically try to clean up temporary files.
  timeoutAddFull (cleanUpTempFiles appStateRef False >> return True) priorityLow 10000

-- Make sure it's visible
  widgetShowAll mainWindow

-- Start the GTK main loop
  mainGUI
