-- MODULE:
--   MapData
--
-- PURPOSE:
--   Data structures for map information
--
-- AUTHOR:
--   Stephen Thompson <stephen@solarflare.org.uk>
--
-- COPYRIGHT:
--   Copyright (C) Stephen Thompson, 2011.
--
--   This file is part of Knights Map Editor.
--
--   Knights Map Editor is free software: you can redistribute it
--   and/or modify it under the terms of the GNU General Public
--   License as published by the Free Software Foundation, either
--   version 3 of the License, or (at your option) any later version.
--
--   Knights Map Editor is distributed in the hope that it will be
--   useful, but WITHOUT ANY WARRANTY; without even the implied
--   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
--   See the GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with Knights Map Editor. If not, see
--   <http://www.gnu.org/licenses/>.


-- NOTE: confusingly, segRooms uses a different coordinate system to the rest of the code:
-- segRooms calls the top left tile of the segment (1,1), while everyone else calls it (0,0).
-- (The idea is that rooms can include the "outer wall" squares which are at 0 and width+1.)
-- Should probably fix this sometime...

module MapData
    (Graphic,
     gfxFilename,    -- Graphic -> String
     gfxColKey,      -- Graphic -> Maybe (Int,Int,Int)
     gfxHandlePos,   -- Graphic -> (Int,Int)
     Tile,           
     tileNumber,     -- Tile -> Int
     tileGfx,        -- Tile -> [Graphic]  (to be drawn in order)
     tileLabel,      -- Tile -> String
     Segment,
     segWidth,       -- Segment -> Int
     segHeight,      -- Segment -> Int
     segGetTile,     -- Segment -> (Int,Int) -> Tile
     segSetTile,     -- Segment -> (Int,Int) -> Tile -> Segment
     segRooms,       -- Segment -> [(Int,Int,Int,Int)]   (same format as in the KConfig i.e. x,y,width,height)
     segSwitches,    -- Segment -> Map (Int,Int) KValue
     segCategory,    -- Segment -> String, can also be used for update
     blankSegment,   -- Tile -> Segment. Fills with the given tile. Size hard coded currently.
     loadSegmentsAndTiles, -- KValue -> ([(String,Segment)], [Tile]) 
                           -- Loads segments & tiles from a "knights_rooms" table.
                           -- Tiles will be returned in increasing tileNumber order.
                           -- Segments will be in the order they were in the KFile. 
     saveSegmentsAndTiles, -- [(String,Segment)] -> [Tile] -> KValue
                           -- opposite of loadSegmentsAndTiles
     prettyPrintKnightsRooms  -- KValue -> String. Call on the result of saveSegmentsAndTiles.
    )

where

import Data.Array
import Data.List
import qualified Data.Map as Map
import Data.Map (Map)
import Data.Maybe
import Data.Ord
import KConfig

-- Graphics.
data Graphic = Graphic { gfxFilename  :: String,
                         gfxColKey    :: Maybe (Int,Int,Int),
                         gfxHandlePos :: (Int,Int)
                       }


-- | Convert a KValue which could be a singleton or a list
-- into a list
readListOrSingle :: KValue -> [KValue]
readListOrSingle val =
    case kList val of 
      Just lst -> lst
      Nothing -> [val]


-- | Reads graphic, depth and label-string from a tile-component (a single entry in the
-- list corresponding to a map-tile)
readTileComponent :: Int -> KValue -> (Maybe Graphic, Int, String)
readTileComponent defaultDepth kValue = 
  let maybeStandardGraphic = kTableLookup "graphic" kValue >>= (Just . graphicFromKValue)
      maybeOpenGraphic = kTableLookup "open_graphic" kValue >>= (Just . graphicFromKValue)
      maybeAnim = kTableLookup "anim" kValue >>= (Just . graphicFromAnim)
      isOpen = maybe False (==1) (kTableLookup "open" kValue >>= kInt)
      depth = maybe defaultDepth id (kTableLookup "depth" kValue >>= kInt)
      label = maybe "" id (kTableLookup "editor_label" kValue >>= kString)
      graphicToUse = if isJust maybeAnim then maybeAnim
                     else if isOpen then maybeOpenGraphic 
                          else maybeStandardGraphic
  in (graphicToUse, depth, label)


-- | Read an entire entry in the tiles list.
-- Returns a list of graphics (in sorted depth order) plus a label-string.
-- The list of graphics will be empty if this is an unused tile-slot.
readTileEntry :: KValue -> ( [Graphic], String )
readTileEntry val = 
    -- gfxAndDepths has type [(Graphic,Int)]
    -- labels has type String
    let (_, gfxAndDepths, labels) = foldl readTileEntry' (0, [], "") (readListOrSingle val)
    in (sortGraphics gfxAndDepths, labels)

-- | This is what combines the tile entries together. Called by readTileEntry.
readTileEntry' :: (Int, [(Graphic,Int)], String) -> KValue -> (Int, [(Graphic,Int)], String)
readTileEntry' (defaultDepth, gfxSoFar, labelSoFar) val =
    case kInt val of
      Just 0 -> (defaultDepth - 10, gfxSoFar, labelSoFar)  -- 0 means we're moving from tiles to items to monsters; reduce the depth by 10
      _ -> let (maybeNewGraphic, newDepth, newLabel) = readTileComponent defaultDepth val
               newGfx = maybe gfxSoFar (\g->(g,newDepth):gfxSoFar) maybeNewGraphic
           in (defaultDepth, newGfx, labelSoFar++newLabel)

-- Sort a gfx list into deepest first order (and then strip off the depths)
sortGraphics :: [(a,Int)] -> [a]
sortGraphics gfx = 
    let sortFn a b = compare (snd b) (snd a)  -- note reversal, so that highest depth will come first. 
        sortedGfx = sortBy sortFn gfx
    in map fst sortedGfx   -- strip off the depths.

debugName :: KValue -> String
debugName kv = case kSymbol kv of
  Just n -> n
  Nothing -> "<unknown>"

-- Convert a KValue to a Graphic
graphicFromKValue :: KValue -> Graphic
graphicFromKValue val = 
    case kList val of 
      Nothing -> error $ "failed to parse graphic: " ++ debugName val
      Just [] -> error $ "failed to parse graphic: empty list: " ++ debugName val
      Just (f:rest) -> 
          case kString f of
            Nothing -> error $ "failed to parse graphic: non-string in lead position: " ++ debugName val
            Just filename -> 
              case rest of
                [] -> Graphic filename Nothing (0,0)
                (r:g:b:[]) -> Graphic filename (Just (fromJust$kInt r, fromJust$kInt g, fromJust$kInt b)) (0,0)
                (r:g:b:x:y:[]) -> Graphic filename (Just (fromJust$kInt r, fromJust$kInt g, fromJust$kInt b)) (fromJust$kInt x, fromJust$kInt y)

-- Convert an Anim KValue into a Graphic
graphicFromAnim :: KValue -> Graphic
graphicFromAnim val =
  case kList val of
    Nothing -> error $ "failed to parse anim: " ++ debugName val
    Just [] -> error $ "failed to parse anim: empty list: " ++ debugName val
    Just [a] -> graphicFromKValue a
    Just (a:b:_) ->
      case kString a of
        Nothing -> graphicFromKValue a  -- First entry isn't a string, so it must be a graphic
        Just _  -> graphicFromKValue b  -- Skip leading "bat" string


-- Tiles
-- (Note a Tile in the editor is not the same as a Tile in the Knights code itself.
-- A Tile here means an entry in "tiles" in the map definitions -- this could map to more than one "Tile" in Knights.)

data Tile = Tile { tileNumber :: Int, 
                   tileGfx    :: [Graphic],   -- in deepest-first order. 
                   tileLabel  :: String,      -- label character(s) to be displayed on top of the tile.
                   origKValue :: KValue       -- for saving back to the file
                 }  

-- Read the entire tile table. Returns a list of Tiles.
readTileTable :: KValue -> [Tile]
readTileTable lst = 
    let lst2 = fromJust $ kList lst
        gfxListsAndLabels = map readTileEntry lst2    -- [([Graphic],String)]
        tiles    = zipWith3 (\(gfx,lab) num kv -> Tile num gfx lab kv) gfxListsAndLabels [1..] lst2
    in filter (\(Tile _ g _ _) -> not (null g)) tiles



-- Segments

data Segment = Segment { segWidth    :: Int,       -- excluding borders
                         segHeight   :: Int,       -- ditto
                         segMap      :: Array (Int,Int) Tile,  -- The map data itself. (also excluding borders!)
                         segRooms    :: [(Int,Int,Int,Int)],  -- x,y,width,height
                         segSwitches :: Map (Int,Int) KValue,
                         segCategory :: String
                       }

segGetTile :: Segment -> (Int,Int) -> Tile
segGetTile seg coords = ((segMap seg)!coords)

segSetTile :: Segment -> (Int,Int) -> Tile -> Segment
segSetTile seg coords newTile =
    let oldMap = segMap seg
        newMap = oldMap // [(coords, newTile)]
    in seg { segMap = newMap }
       
       
-- | Creates a "blank" segment, suitable for adding into the map
-- Parameter = tile to fill with.
-- NOTE: Size is hard coded currently.
blankSegment :: Tile -> Segment
blankSegment floorTile = 
  Segment { segWidth = w,
            segHeight = h,
            segMap = array ((0,0),(w-1,h-1)) [((x,y),floorTile) | x <- [0..w-1], y <- [0..h-1]],
            segRooms = [],
            segSwitches = Map.empty,
            segCategory = "" }
    where w = 12
          h = 12



-- Load a segment from its representation as a KTable.
-- Inputs: * a function telling us how to convert tile numbers to Tiles
--         * a KTable representing the segment
-- Output: a Segment. 

readSegment :: (Int -> Tile) -> KValue -> (String,Segment)
readSegment getTile tab = 
    let width  = fromJust $ kInt $ fromJust $ kTableLookup "width" tab
        height = fromJust $ kInt $ fromJust $ kTableLookup "height" tab
        name = fromJust $ kString $ fromJust $ kTableLookup "name" tab
        tileList = map (getTile . fromJust . kInt) $ fromJust $ kList $ fromJust $ kTableLookup "data" tab 
        mapAssocList = snd $ mapAccumL f (0,0) tileList          -- convert list of tiles into association list from coords to tiles
            where f coord tile = (nextCoord coord, (coord,tile))
                  nextCoord (x,y) | (x+1 < width) = (x+1, y)
                  nextCoord (x,y) | otherwise     = (0, y+1)
        mapData = array ((0,0),(width-1,height-1)) mapAssocList
        roomList = kTableLookup "rooms" tab
        rooms = case roomList of 
                  Nothing -> []
                  Just rl -> 
                      let lstOfLstOfKvals = map (fromJust.kList) ((fromJust.kList) rl)
                          lstOfLstOfInt = map (map (fromJust.kInt)) lstOfLstOfKvals
                      in map (\[a,b,c,d]->(a,b,c,d)) lstOfLstOfInt
        switchList = kTableLookup "switches" tab
        switches = case switchList of 
                     Nothing -> []
                     Just sl -> fromJust $ kList sl
        maybeCategory = kTableLookup "category" tab
        category = case maybeCategory of
                     Nothing -> ""
                     Just c -> fromJust $ kString c 
    in (name, 
        Segment { segWidth = width,
                  segHeight = height, 
                  segMap = mapData,
                  segRooms = rooms,
                  segSwitches = convertSwitchList switches,
                  segCategory = category
                }
       )

writeRoom (a,b,c,d) = makeKList $ map makeKInt $ [a,b,c,d]

writeRoomData :: [(Int,Int,Int,Int)] -> KValue
writeRoomData = makeKList . map writeRoom

writeSwitch :: ((Int,Int),KValue) -> KValue
writeSwitch ((x,y),kv) = makeKList [makeKInt x, makeKInt y, kv]

writeSegmentData :: Segment -> KValue
writeSegmentData seg = 
  let coords = [(x,y) | y <- [0..(segHeight seg)-1], x <- [0..(segWidth seg)-1]]
  in makeKList $ map (\c -> makeKInt $ tileNumber $ (segMap seg)!c) coords

writeSegment :: (String,Segment) -> KValue
writeSegment (segName, seg) =
    makeKTable (Map.fromList (makeSegmentList segName seg ++ makeCategoryList seg))
    
makeSegmentList segName seg = [("width", makeKInt $ segWidth seg)
                              ,("height", makeKInt $ segHeight seg)
                              ,("name", makeKString $ segName)
                              ,("data", writeSegmentData seg)
                              ,("rooms", writeRoomData $ segRooms seg)
                              ,("switches", makeKList $ map writeSwitch $ sortByCoord $ Map.toList $ segSwitches seg)]
-- note: bat_placement_tile is hard coded to 73 for guarded exits, unset otherwise.
makeCategoryList seg = 
  case segCategory seg of 
    "" -> []
    x | x == "guarded_exit" -> [("category", makeKString x), ("bat_placement_tile", makeKInt 73)]
    x -> [("category", makeKString x)]


-- Sort the switches list, left to right then top to bottom.
-- Not strictly necessary, but this makes the list be in the same order as in kconvert.py,
-- which makes comparing the results easier.
sortByCoord :: [((Int,Int),a)] -> [((Int,Int),a)]
sortByCoord = sortBy (comparing ( \((x,y),_) -> (y,x) ) )

convertSwitchList :: [KValue] -> Map (Int,Int) KValue
convertSwitchList = Map.fromList . (map f)
    where f kv = case kList kv of
                   Just [x,y,a] -> 
                       case (kInt x, kInt y) of
                         (Just xx, Just yy) -> ((xx,yy),a)
                         _ -> error $ (show kv) ++ " (1) problem in convertSwitchList"
                   _ -> error $ (show kv) ++ " (2) problem in convertSwitchList"


-- Load all Tiles and Segments from a "knights_rooms" table. 
loadSegmentsAndTiles :: KValue -> ([(String,Segment)], [Tile])
loadSegmentsAndTiles kval =
    let Just inputTiles    = kTableLookup "tiles" kval
        Just inputSegments = kTableLookup "segments" kval
        tiles    = readTileTable inputTiles
        segments = map (readSegment lookupTile) (fromJust $ kList inputSegments)
            where lookupTile n = case find (\t->(tileNumber t == n)) tiles of 
                                   Just t' -> t'
                                   Nothing -> error "tile not found in tile list!"
    in (segments, tiles)


saveSegmentsAndTiles :: [(String,Segment)] -> [Tile] -> KValue
saveSegmentsAndTiles segments tiles = 
  let kListTiles = makeKList (makeTileList tiles)
      kListSegments = makeKList (map writeSegment segments)
  in makeKTable (Map.fromList [("tiles", kListTiles), ("segments", kListSegments)])

makeTileList :: [Tile] -> [KValue]
makeTileList tiles = 
    let tileNumbers = map tileNumber tiles
        diffs = 1 : (zipWith (-) (tail tileNumbers) tileNumbers)
        diffsWithTiles :: [(Int,Tile)]
        diffsWithTiles = zip (map (\x->x-1) diffs) tiles
    in concatMap (\(d,t) -> (take d $ repeat $ makeKInt 0) ++ [origKValue t]) diffsWithTiles


prettyPrintKnightsRooms :: KValue -> String
prettyPrintKnightsRooms krooms =
    let (Just tiles) = kTableLookup "tiles" krooms
        (Just segments) = kTableLookup "segments" krooms
    in "# Knights rooms file, saved by Knights Map Editor\n\n"
           ++ "include \"tiles.txt\"\n\n"
           ++ "knights_rooms = {\n\n" ++ (prettyPrintTiles tiles) 
           ++ "\n    segments = [" ++ (prettyPrintSegments segments)
           ++ "    ]\n}\n"

prettyPrintSegments :: KValue -> String
prettyPrintSegments = (concatMap prettyPrintSegment) . fromJust . kList

prettyPrintSegment :: KValue -> String
prettyPrintSegment seg =
    let widthStr = printFromKTable "width" seg
        heightStr = printFromKTable "height" seg
        categoryStr = printFromKTable "category" seg
        batPlacementStr = printFromKTable "bat_placement_tile" seg
        nameStr = printFromKTable "name" seg
        dataList = fromJust $ kList $ fromJust $ kTableLookup "data" seg
        switchesList = fromJust $ kList $ fromJust $ kTableLookup "switches" seg
        roomsStr = show (fromJust $ kTableLookup "rooms" seg)
    in "{\n" ++ widthStr ++ heightStr ++ categoryStr ++ batPlacementStr
       ++ nameStr ++ " data = [\n" ++ (prettyPrintDataList dataList) ++ " ]\n"
       ++ (prettyPrintSwitches switchesList) ++ " rooms = " ++ roomsStr ++ "\n}\n\n"

printFromKTable key tbl = 
  case kTableLookup key tbl of
    Nothing -> ""
    Just kv -> " " ++ key ++ " = " ++ (show kv) ++ "\n"

prettyPrintSwitches :: [KValue] -> String
prettyPrintSwitches [] = ""
prettyPrintSwitches xs = " switches = [\n" ++ (unlines $ map prettyPrintSwitch xs) ++ " ]\n"

prettyPrintSwitch :: KValue -> String
prettyPrintSwitch x = "   " ++ showSwitch x

-- Show a KValue but some changes to be consistent with how kconvert.py formats the output
showSwitch :: KValue -> String
showSwitch kv = showSwitchWith normalLst kv

showSwitchWith :: ([KValue]->String) -> KValue -> String
showSwitchWith f kv = case kSymbol kv of
                        Just s -> s
                        Nothing -> case kList kv of
                                     Nothing -> case kDirective kv of 
                                                  Nothing -> show kv
                                                  Just (str,kv2) -> str ++ "(" ++ (showSwitchWith directiveLst kv2) ++ ")"
                                     Just xs -> f xs
normalLst xs = "[" ++ (removeFinal $ concatMap (\x -> (showSwitchWith normalLst x) ++ " ") xs) ++ "]"
directiveLst xs = (removeFinal $ concatMap (\x -> (showSwitchWith normalLst x) ++ ",") xs)

removeFinal :: [a] -> [a]
removeFinal [] = []
removeFinal xs = init xs


-- Split a list into groups of n.
-- Probably not the most elegant way of doing this, but it works.
getGroups :: Int -> [a] -> [[a]]
getGroups n xs = 
    let (a,b) = splitAt n xs
    in case b of 
         [] -> [a]
         bs -> a : getGroups n bs

prettyPrintDataList :: [KValue] -> String
prettyPrintDataList kvals =
    let strings = map (\x -> (show x) ++ " ") kvals
        groups = getGroups 12 strings
        lines = map (\g -> (concat g) ++ "\n") groups
    in concat lines

prettyPrintTiles :: KValue -> String
prettyPrintTiles kv = 
  "    tiles = [\n" ++ concatMap prettyPrintTile (fromJust $ kList kv) ++ "    ]\n"

prettyPrintTile :: KValue -> String
prettyPrintTile kv = "        " ++ show kv ++ "\n"
